# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# Copyright (C) 2005 Laurent A.V. Szyster
# 
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# 
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

"http://laurentszyster.be/blog/select_trigger/"

import sys, os, socket, thread

from allegra import (
	netstring, prompt, loginfo, 
        async_loop, finalization, async_core
        )


class Trigger (async_core.Dispatcher):

        "Thunk back safely from threads into the asynchronous loop"
        
        def __repr__ (self):
                return 'trigger id="%x"' % id (self)

        def readable (self):
                "a Trigger is allways readable"
                return True

        def writable (self):
                "a Trigger is never writable"
                return False

        def handle_connect (self):
                "pass on connect"
                pass

        def handle_read (self):
                "try to call all thunked, log all exceptions' traceback"
                try:
                        self.recv (8192)
                except socket.error:
                        return
                
                self.lock.acquire ()
                try:
                        thunks = self.thunks
                        self.thunks = []
                finally:
                        self.lock.release ()
                for thunk in thunks:
                        try:
                                thunk[0] (*thunk[1])
                        except:
                                self.loginfo_traceback ()
                        

if os.name == 'posix':

        def posix_trigger_init (self):
                "use a POSIX pipe to connect a pair of file descriptors"
                self.select_triggers = 0
                fd, self.trigger = os.pipe ()
                self.set_file (fd)
                self.lock = thread.allocate_lock ()
                self.thunks = []
                assert None == self.log ('open', 'debug')

        def posix_trigger_pull (self, thunk):
                "acquire the trigger's lock, thunk and pull"
                self.lock.acquire ()
                try:
                        self.thunks.append (thunk)
                finally:
                        self.lock.release ()
                os.write (self.trigger, 'x')
        
        def posix_trigger_close (self):
                "close the trigger"
                async_core.Dispatcher.close (self)
                os.close (self.trigger)

        Trigger.__init__ = posix_trigger_init
        Trigger.__call__ = posix_trigger_pull
        Trigger.close = posix_trigger_close
                                
elif os.name == 'nt':

        def win32_trigger_init (self):
                "get a pair of Win32 connected sockets"
                self.select_triggers = 0
                a = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
                w = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
                a.bind (('127.9.9.9', 19999))
                a.listen (1)
                w.setblocking (0)
                try:
                        w.connect (('127.9.9.9', 19999))
                except:
                        pass
                conn, addr = a.accept ()
                a.close ()
                w.setblocking (1)
                self.trigger = w
                self.set_connection (conn, addr)
                self.lock = thread.allocate_lock ()
                self.thunks = []
                assert None == self.log ('open', 'debug')

        def win32_trigger_pull (self, thunk):
                "acquire the trigger's lock, thunk and pull"
                self.lock.acquire ()
                try:
                        self.thunks.append (thunk)
                finally:
                        self.lock.release ()
                self.trigger.send ('x')

        def win32_trigger_close (self):
                "close the trigger"
                async_core.Dispatcher.close (self)
                self.trigger.close ()

        Trigger.__init__ = win32_trigger_init
        Trigger.__call__ = win32_trigger_pull
        Trigger.close = win32_trigger_close

else:
	raise ImportError ('OS "%s" not supported, sorry :-(' % os.name)


class Select_trigger (loginfo.Loginfo, finalization.Finalization):
	
	"""A base class that implements the select_trigger interface
	
		select_trigger ((function, args))
	
	to thunk function and method calls from one thread into the main
	asynchronous loop. Select_trigger implements thread-safe and
	practical loginfo interfaces:
		
		select_trigger_log (data, info=None)
		
	to log information, and
		
		select_trigger_traceback ()
		
	to log traceback asynchronously from a distinct thread."""

	select_trigger = None

	def __init__ (self):
                "maybe open a new Trigger, increase its reference count"
		if self.select_trigger == None:
			Select_trigger.select_trigger = Trigger ()
		self.select_trigger.select_triggers += 1
		
	def __repr__ (self):
		return 'select-trigger id="%x"' % id (self)
		
	def select_trigger_log (self, data, info=None):
		"log asynchronously via the select trigger"
		self.select_trigger ((self.log, (data, info)))
		
	def select_trigger_traceback (self):
		"return a compact traceback tuple and log asynchronously"
		ctb = prompt.compact_traceback ()
		self.select_trigger ((
                        self.loginfo_log,
			loginfo.traceback_encode (ctb)
			))
		return ctb

	def finalization (self, finalized):
		"decrease the Trigger's reference count, maybe close it"
                trigger = self.select_trigger
		trigger.select_triggers -= 1
		if trigger.select_triggers == 0:
			trigger ((trigger.handle_close, ()))
			Select_trigger.select_trigger = None
		self.select_trigger = None


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/thread_loop/"

import threading, collections

from allegra import netstring, loginfo, finalization, select_trigger


class Trunked_deque (object):
        
        "a deque implementation to trunk protected deque safely"

        def __len__ (self):
                "return 1, a trunked deque has allways one item" 
                return 1
                
        def __getitem__ (self, index):
                "return None or raise IndexError"
                if index == 0:
                        return None
                        
                raise IndexError, 'trunked deque'
                
        def append (self, item): 
                "drop any item appended to the deque"
                pass

        def popleft (self): 
                "return None, the closing item for a deque consumer"
                return None
        
        pop = popleft
        
        appendleft = append


class Protected_deque (object):

        "a thread-safe wrapper for a deque"

	def __init__ (self, queue=None):
		self.deque = collections.deque (queue or [])
		self.mon = threading.RLock ()
		self.cv = threading.Condition (self.mon)

        def __repr__ (self):
                "return a safe netstring representation of the deque"
                r = []
                try:
                        self.cv.acquire ()
                        l = len (self.deque)
                        for item in self.deque:
                                try:
                                        r.append ('%r' % (item,))
                                except:
                                        r.append (
                                                'item id="%x"' % id (item)
                                                )
                finally:
                        self.cv.release ()
                return netstring.encode ((
                        'protected_deque queued="%d"' % l,
                        netstring.encode (r)
                        ))
        
	def __len__ (self):
                "return the queue's length"
		try:
			self.cv.acquire ()
			l = len (self.deque)
		finally:
			self.cv.release ()
		return l

        def __getitem__ (self, index):
                "return the queue's length"
                try:
                        self.cv.acquire ()
                        return self.deque[index]
                                
                finally:
                        self.cv.release ()
                
        def append (self, item):
                "push an item at the end the deque and notify"
                try:
                        self.cv.acquire ()
                        self.deque.append (item)
                        self.cv.notify ()
                finally:
                        self.cv.release ()
        
        __call__ = append
                
        def popleft (self):
                "wait for a first item in the deque and pop it"
                try:
                        self.cv.acquire ()
                        while len (self.deque) == 0:
                                self.cv.wait ()
                        item = self.deque.popleft ()
                finally:
                        self.cv.release ()
                return item

        def appendleft (self, item):
                "push an item of items in front of the deque"
                try:
                        self.cv.acquire ()
                        self.deque.appendleft (item)
                        self.cv.notify ()
                finally:
                        self.cv.release ()

        def pop (self):
                "wait for a last item in the deque and pop it"
                try:
                        self.cv.acquire ()
                        while len (self.deque) == 0:
                                self.cv.wait ()
                        item = self.deque.pop ()
                finally:
                        self.cv.release ()
                return item

        def trunk (self):
                """Replace the deque with a trunked deque implementation
                and return the replaced deque instance."""
                try:
                        self.cv.acquire ()
                        trunked = self.deque
                        self.deque = Trunked_deque ()
                finally:
                        self.cv.release ()
                return trunked
                #
                # In effect, trunking implies closing a running thread loop 
                # and dropping any item queued thereafter, which is precisely 
                # The Right Thing To Do when a thread loop queue is stopped: 
                # prevent accessors to push  references that won't be popped 
                # out and leak.                
                

class Thread_loop (threading.Thread, select_trigger.Select_trigger):

        "a thread loop, with thread-safe asynchronous logging"

	def __init__ (self, queue=None):
		self.thread_loop_queue = queue or Protected_deque ()
		select_trigger.Select_trigger.__init__ (self)
		threading.Thread.__init__ (self)
		self.setDaemon (1)

	def __repr__ (self):
		return 'thread-loop id="%x"' % id (self)

	def run (self):
                """The Thread Loop
                
                If thread_loop_init() is True call queued instance until
                None is popped or and exception is raised and not catched
                by thread_loop_throw. Finally, if thread_loop_delete() is
                True, trunk the thread loop queue.
                """
		if self.thread_loop_init ():
                        next = self.thread_loop_queue.popleft # ? maybe safer
			while True:
				queued = next () # ... sure faster
				if queued == None:
					break

				try:
				        queued[0] (*queued[1])
				except:
					if self.thread_loop_throw ():
                                                del queued
                                                break
                                        
                                else:
                                        del queued
                        #
                        # note that I make sure to delete the tuple which
                        # would otherwise hold a reference to the method and
                        # arguments of the call threaded, preventing garbage
                        # collection hence finalization and was the source
                        # of subtle bugs ...
                        #
		if self.thread_loop_delete ():
                        trunked = self.thread_loop_queue.trunk ()
                        if trunked:
                                assert None == self.select_trigger_log (
                                        netstring.encode ([
                                                '%r' % (i,) for i in trunked
                                                ]), 'debug'
                                        )
		#
		# ... continue with the Select_trigger.finalization, unless
                # there are circular references for this instance, caveat!

        def thread_loop (self, queued):
                "assert debug log and push a simple callable in the queue"
                assert None == self.log ('%r %r' % queued, 'queued')
                self.thread_loop_queue (queued)

        def thread_loop_stop (self):
                "assert debug log and push the stop item None in the queue"
                assert None == self.log ('stop-when-done', 'debug')
                self.thread_loop_queue (None)

	def thread_loop_init (self):
                "return True, assert a debug log of the thread loop start"
		assert None == self.log ('start', 'debug')
		return True
                
        def thread_loop_throw (self):
                "return False, log a compact traceback via the select trigger"
                self.select_trigger_traceback ()
                return False

        def thread_loop_delete (self):
                "return True, assert a debug log of the thread loop start"
                assert None == self.log ('stop', 'debug')
                return True


class Synchronizer (loginfo.Loginfo):

        def __init__ (self, size=2):
                self.synchronizer_size = size
                self.synchronized_thread_loops = []
                self.synchronized_instance_count = []
                self.synchronized_count = 0

        def __repr__ (self):
                return 'synchronizer pid="%x" count="%d"' % (
                        id (self), self.synchronized_count
                        )

        def synchronizer_append (self):
                assert None == self.log (
                        'append %d' % len (self.synchronized_thread_loops), 
                        'synchronizer'
                        )
                t = Thread_loop ()
                t.thread_loop_queue.synchronizer_index = len (
                        self.synchronized_thread_loops
                        )
                self.synchronized_thread_loops.append (t)
                self.synchronized_instance_count.append (0)
                t.start ()
                
        def synchronize (self, instance):
                assert not hasattr (instance, 'synchronized')
                if self.synchronized_count == len (
                        self.synchronized_thread_loops
                        ) < self.synchronizer_size:
                        self.synchronizer_append ()
                index = self.synchronized_instance_count.index (
                        min (self.synchronized_instance_count)
                        )
                t = self.synchronized_thread_loops[index]
                instance.synchronized = t.thread_loop_queue
                instance.select_trigger = t.select_trigger 
                self.synchronized_instance_count[index] += 1
                self.synchronized_count += 1
                assert None == self.log ('%r' % instance, 'synchronized')

        def desynchronize (self, instance):
                assert hasattr (instance, 'synchronized')
                i = instance.synchronized.synchronizer_index
                count = self.synchronized_instance_count[i]
                self.synchronized_count += -1
                self.synchronized_instance_count[i] += -1
                instance.select_trigger = instance.synchronized = None
                if self.synchronized_count == 0:
                        assert None == self.log ('stop %d threads' % len (
                                self.synchronized_thread_loops
                                ), 'synchronizer')
                        for t in self.synchronized_thread_loops:
                                t.thread_loop_queue (None)
                        self.synchronized_thread_loops = []
                assert None == self.log ('%r' % instance, 'desynchronized')


def synchronize (instance):
        if instance.synchronizer == None:
                instance.__class__.synchronizer = Synchronizer (
                        instance.synchronizer_size
                        )
        instance.synchronizer.synchronize (instance)


def desynchronize (instance):
        instance.synchronizer.desynchronize (instance)


def synchronized (instance):
        assert isinstance (instance, finalization.Finalization) 
        if instance.synchronizer == None:
                instance.__class__.synchronizer = Synchronizer (
                        instance.synchronizer_size
                        )
        instance.synchronizer.synchronize (instance)
        instance.finalization = instance.synchronizer.desynchronize

        
# Notes about the Synchronizer
#
# The purpose is to but deliver non-blocking interfaces to synchronous API.
#
# The synchronizer is an resizable array of thread loop queues. Synchronized
# instances are attached to one of these queues. When a synchronized instance
# is finalized, that reference is released and the array is notified. When no
# more instance is attached to a thread loop queue, its thread exits. If the
# limit set on the array size is not reached, a new thread loop is created for
# each new synchronized instance. The default limit is set to 4.
#
# This interface is purely asynchronous: methods synchronized should be able
# to access the select_trigger to manipulate the Synchronizer, or more
# mundanely to push data to asynchat ...
#
#
# Limits
#
# There is no easy way to prevent an instance to stall its thread loop queue
# and all the other instances methods synchronized to it. The only practical
# algorithm to detect a stalled method (and "fix" it), is to set a limit on
# the size of the synchronized queue and when that limit is reached to replace
# the stalled thread loop by a new one. However, this would leave the stalled
# thread to hang forever if the stalling method is running amok or blocking
# forever too. Setting a timeout on each synchronized method is impossible
# since there is no way to infer reliably a maximum execution time, certainly
# in such case of concurrent processes.
#
# Basicaly, there is no practical and effective way to fix a thread broken by
# an infinite loop or a stalled-forever wait state. So, this implementation
# does not even attempt to correct the effects of such bugs on the other
# synchronized instance methods.
#
#
# Beware!
#
# Synchronized methods must be tested separately. Yet it is trivial, because
# you may either test them asynchronously from within an async_loop host or,
# since they are synchronous, directly from the Python prompt.
#
# My advice is to use synchronized method in two cases. Either you don't want
# to learn asynchronous programming (don't have time for that). Or you know
# how, but need to access a blocking API that happens to be thread safe and
# releases the Python GIL.
#
# For instance:
#
#         os.open (...).read ()
#
# or
#
#        bsddb.db.DB ().open (...)
#
# may be blocking and should be synchronized.
        


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/sync_stdio/"

import sys

from allegra import prompt, loginfo, async_loop, finalization, thread_loop


logger = loginfo.logger


class Sync_stdio (thread_loop.Thread_loop):

	def __init__ (self):
                self.async_loop_catch = async_loop._catched
                async_loop._catched = self.async_prompt_catch
                self.sync_stdout = logger.loginfo_stdout
                self.sync_stderr = logger.loginfo_stderr
                logger.loginfo_stdout = self.async_stdout
                logger.loginfo_stderr = self.async_stderr
		thread_loop.Thread_loop.__init__ (self)

	def __repr__ (self):
		return 'sync-stdio'

	def async_stdout (self, data):
		self.thread_loop_queue ((self.sync_stdout, (data,)))

	def async_stderr (self, data):
		self.thread_loop_queue ((self.sync_stderr, (data,)))

	def async_stdio_stop (self):
                async_loop._catched = self.async_loop_catch
                self.async_loop_catch = None
                logger.loginfo_stdout = self.sync_stdout
                logger.loginfo_stderr = self.sync_stderr
		try:
			del self.log
		except:
			pass
		self.thread_loop_queue (None)
                return True

        async_prompt_catch = async_stdio_stop

        def thread_loop_delete (self):
                assert None == self.select_trigger_log (
                        'stdio-stop', 'debug'
                        )
                # del self.sync_stdout, self.sync_stderr
                return True
                

class Sync_stdoe (Sync_stdio):

	def __repr__ (self):
		return 'sync-stdoe'

	def thread_loop_init (self):
                self.select_trigger_log (
                        'Press CTRL+C to stop synchronous I/O', 'info'
                        )
		self.thread_loop_queue ((self.sync_stdin_close, ()))
		return True

	def sync_stdin_close (self):
		sys.stdin.close ()
		assert None == self.select_trigger_log (
			'stdin_close', 'debug'
			)


class Sync_prompt (Sync_stdio):

        sync_prompt_prefix = '>>> '
        sync_prompt_comment = '#'
        sync_prompt_ready = not __debug__
        
	def thread_loop_init (self):
                if self.sync_prompt_ready:
                        self.thread_loop_queue ((self.sync_stdin, ()))
                else:
                        self.select_trigger_log (
                                'press CTRL+C to open and close the console', 
                                'info'
                                )
		return True
		
	def async_prompt_catch (self):
		if self.sync_prompt_ready:
			self.thread_loop_queue ((
				self.sync_stderr, ('[CTRL+C]\n',)
				))
                        self.sync_prompt_ready = False
		else:
			self.thread_loop_queue ((self.sync_stdin, ()))
			self.sync_prompt_ready = True
		return True

	def sync_stdin (self):
	        self.sync_stderr (self.sync_prompt_prefix)
                line = sys.stdin.readline ()
		if line == '':
                        if sys.stdin.closed:
                                self.select_trigger ((
                                        self.async_stdio_stop, ()
                                        ))
                        else:
                                self.select_trigger ((self.async_prompt, ()))
                elif line == '\n':
			self.select_trigger ((self.async_prompt, ()))
                else:
        		self.select_trigger ((
                                self.async_readline, (line[:-1],)
                                ))
                
        def sync_stdin_script (self):
                line = sys.stdin.readline ()
                if line == '':
                        self.select_trigger ((self.async_prompt, ()))
                elif line == '\n' or line.startswith (
                        self.sync_prompt_comment
                        ):
                        self.thread_loop_queue ((self.sync_stdin, ()))
                else:
                        self.select_trigger ((
                                self.async_readline, (line[:-1],)
                                ))
                                
        def async_prompt (self):
                self.sync_prompt_ready = False
                
	def async_readline (self, line):
		assert None == self.log (line)
		self.thread_loop_queue ((self.sync_stdin, ()))


class Python_prompt (Sync_prompt):

	def __init__ (self, env=None, info=None):
                self.loginfo_info = info
		self.python_prompt_env = env or {'prompt': self}
                loginfo.log ('Python %s on %s' % (
                        sys.version, sys.platform
                        ), info)
		Sync_prompt.__init__ (self)

	def __repr__ (self):
		return 'python-prompt id="%x"' % id (self)

	def async_readline (self, line):
		method, result = prompt.python_prompt (
			line, self.python_prompt_env
			)
		if method == 'excp':
			self.loginfo_traceback (result)
		elif result != None:
                        if __debug__:
			        self.async_stderr ('%r\n' % (result,))
                        else:
                                self.select_trigger ((loginfo.log, (
                                        '%r' % (result, ), self.loginfo_info
                                        )))
                if self.async_loop_catch != None: 
		        self.thread_loop_queue ((self.sync_stdin, ()))

	def async_stdio_stop (self):
		self.python_prompt_env = None # break circular reference
		return Sync_prompt.async_stdio_stop (self)
		

if __name__ == '__main__':
        import sys
        info = None # log prompt results to STDOUT by default
        if '-d' in sys.argv:
                sys.argv.remove ('-d')
                loginfo.Logger.log = loginfo.Logger.loginfo_netlines
        elif not __debug__:
                Python_prompt.sync_stdin = Sync_prompt.sync_stdin_script
                info = 'prompt'
	assert None == loginfo.log (
		'Allegra Prompt'
		' - Copyright 2005 Laurent A.V. Szyster'
		' | Copyleft GPL 2.0', 'info'
		)
        Python_prompt (None, info).start ()
        async_loop.dispatch ()
        assert None == finalization.collect ()


# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#
# ... and, for the little added, also ...
#
#
# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_core/"

import exceptions, sys, os, time, socket, collections

from errno import (
        EALREADY, EINPROGRESS, EWOULDBLOCK, ECONNRESET, 
        ENOTCONN, ESHUTDOWN, EINTR, EISCONN
        )

from allegra import loginfo, async_loop, finalization


class Dispatcher (loginfo.Loginfo, finalization.Finalization):
        
        connected = accepting = closing = False
        socket = addr = family_and_type = _fileno = None

        def create_socket (self, family, type):
                "create a socket and add the dispatcher to the I/O map"
                self.family_and_type = family, type
                self.socket = socket.socket (family, type)
                self.socket.setblocking (0)
                self._fileno = self.socket.fileno ()
                self.add_channel ()

        def set_connection (self, conn, addr):
                "set a connected socket and add the dispatcher to the I/O map"
                conn.setblocking (0)
                self.socket = conn
                self.addr = addr
                self._fileno = conn.fileno ()
                self.add_channel ()
                self.connected = True

        def set_reuse_addr (self):
                "try to re-use a server port if possible"
                try:
                        self.socket.setsockopt (
                                socket.SOL_SOCKET, 
                                socket.SO_REUSEADDR,
                                self.socket.getsockopt (
                                        socket.SOL_SOCKET, 
                                        socket.SO_REUSEADDR
                                        ) | 1
                                )
                except socket.error:
                        pass

        def listen (self, num):
                "listen and set the dispatcher's accepting state"
                self.accepting = True
                if os.name == 'nt' and num > 5:
                        num = 1
                return self.socket.listen (num)

        def bind (self, addr):
                "bind to addr and set the dispatcher's addr property"
                self.addr = addr
                return self.socket.bind (addr)

        def connect (self, address):
                "try to connect and set the dispatcher's connected state"
                err = self.socket.connect_ex (address)
                if err in (EINPROGRESS, EALREADY, EWOULDBLOCK):
                        return
                        
                if err in (0, EISCONN):
                        self.addr = address
                        self.connected = True
                        self.handle_connect ()
                else:
                        raise socket.error, err

        def accept (self):
                "try to accept a connection"
                try:
                        conn, addr = self.socket.accept()
                        return conn, addr
                        
                except socket.error, why:
                        if why[0] == EWOULDBLOCK:
                                pass
                        else:
                                raise
                        
        def close (self):
                "close the socket and remove the dispatcher from the I/O map"
                try:
                        self.socket.close ()
                except:
                        pass # closing a
                self.socket = None 
                self.del_channel ()
                self.connected = False
                self.closing = True # == (self.socket == None)
                assert None == self.log ('close', 'debug')

        # The transport API for stream and datagram sockets

        def send (self, data):
                "try to send data through a stream socket"
                try:
                        result = self.socket.send (data)
                        return result
                        
                except socket.error, why:
                        if why[0] == EWOULDBLOCK:
                                return 0
                                
                        else:
                                raise
                                
                        return 0

        def recv (self, buffer_size):
                "try to receive bytes from a stream socket or handle close"
                try:
                        data = self.socket.recv (buffer_size)
                        if not data:
                                # a closed connection is indicated by signaling
                                # a read condition, and having recv() return 0.
                                self.handle_close()
                                return ''
                                
                        else:
                                return data
                        
                except MemoryError:
                        # according to Sam Rushing, this is a place where
                        # MemoryError tend to be raised by Medusa. the rational 
                        # is that under high load, like a DDoS (or the /.
                        # effect :-), recv is the function that will be called 
                        # most *and* allocate the more memory.
                        #
                        sys.exit ("Out of Memory!") # do not even try to log!
                        
                except socket.error, why:
                        # winsock sometimes throws ENOTCONN
                        if why[0] in [ECONNRESET, ENOTCONN, ESHUTDOWN]:
                                self.handle_close()
                                return ''
                                
                        else:
                                raise

        def sendto (self, data, peer):
                "try to send data through a datagram socket"
                try:
                        return self.socket.sendto (data, peer)
                
                except socket.error, why:
                        if why[0] == EWOULDBLOCK:
                                return 0
                                
                        else:
                                raise
                                
                        return 0

        def recvfrom (self, datagram_size):
                "try to receive from a datagram socket, maybe handle close"
                try:
                        return self.socket.recvfrom (datagram_size)

                except socket.error, why:
                        if why[0] in [ECONNRESET, ENOTCONN, ESHUTDOWN]:
                                self.handle_close()
                                return '', None
                                
                        else:
                                raise

        # The "iner" API applied in async_loop

        async_map = async_loop._dispatched
    
        def add_channel (self):
                "add the dispatcher to the asynchronous I/O map"
                self.async_map[self._fileno] = self

        def del_channel (self):
                "removes the dispatcher from the asynchronous I/O map"
                fd = self._fileno
                try:
                        del self.async_map[fd]
                except KeyError:
                        pass

        def readable (self):
                "predicate for inclusion as readable in the poll loop"
                return True

        def writable (self):
                "predicate for inclusion as writable in the poll loop"
                return True

        def handle_read_event (self):
                "articulate read event as accept, connect or read."
                if self.accepting:
                        # for an accepting socket, getting a read implies
                        # that we are connected
                        #
                        # TODO: is this actually usefull? I mean, a listener
                        #       is not a stream connection, not really ...
                        #
                        if not self.connected:
                                self.connected = True
                        self.handle_accept ()
                elif not self.connected:
                        self.handle_connect ()
                        self.connected = True
                        self.handle_read ()
                else:
                        self.handle_read ()

        def handle_write_event (self):
                "articulate write event as connect or write."
                if not self.connected:
                        # getting a write implies that we are connected
                        self.handle_connect ()
                        self.connected = True
                self.handle_write ()
                
        # The protocol API for stream and datagram sockets

        def handle_error (self):
                "log a traceback and close, or raise SystemExit again"
                t, v = sys.exc_info ()[:2]
                if t is SystemExit:
                        raise t, v

                self.loginfo_traceback ()
                self.close () # self.handle_close () ... or nothing?

        def handle_close (self):
                self.close ()

        def handle_read (self):
                "to subclass: assert unhandled read event debug log"
                assert None == self.log ('unhandled read event', 'debug')

        def handle_write (self):
                "to subclass: assert unhandled write event debug log"
                assert None == self.log ('unhandled write event', 'debug')

        def handle_connect (self):
                "to subclass: assert unhandled connect event debug log"
                assert None == self.log ('unhandled connect event', 'debug')

        def handle_accept (self):
                "to subclass: assert unhandled accept event debug log"
                assert None == self.log ('unhandled accept event', 'debug')
                
        # and finaly ...

        def finalization (self, finalized):
                "assert debug log of the instance finalization"
                assert None == self.log ('finalized', 'debug')


# Asynchronous File I/O: UNIX pipe and stdio only
#
# What follows is the original comments by Sam Rushing:
#
# After a little research (reading man pages on various unixen, and
# digging through the linux kernel), I've determined that select()
# isn't meant for doing doing asynchronous file i/o.
# Heartening, though - reading linux/mm/filemap.c shows that linux
# supports asynchronous read-ahead.  So _MOST_ of the time, the data
# will be sitting in memory for us already when we go to read it.
#
# What other OS's (besides NT) support async file i/o?  [VMS?]
#
# Regardless, this is useful for pipes, and stdin/stdout...

if os.name == 'posix':
        import fcntl

        class File_wrapper (object):

                "wrap a file with enough of a socket like interface"

                def __init__ (self, fd):
                        self.fd = fd
        
                def recv (self, *args):
                        return apply (os.read, (self.fd, ) + args)
        
                def send (self, *args):
                        return apply (os.write, (self.fd, ) + args)
        
                read = recv
                write = send
        
                def close (self):
                        return os.close (self.fd)
        
                def fileno (self):
                        return self.fd
                        
        def set_file (self, fd):
                "set a file descriptor and add the dispatcher (POSIX only)"
                flags = fcntl.fcntl (fd, fcntl.F_GETFL, 0) | os.O_NONBLOCK
                fcntl.fcntl (fd, fcntl.F_SETFL, flags)
                self._fileno = fd
                self.socket = File_wrapper (fd)
                self.add_channel ()
                self.connected = True

        Dispatcher.set_file = set_file


# Conveniences

class Dispatcher_with_send (Dispatcher):
        
        ac_out_buffer_size = 1 << 14 # sweet sixteen kilobytes

        def __init__ (self):
                self.ac_out_buffer = ''
                
        def writable (self):
                "writable when there is output buffered or queued"
                return not (
                        (self.ac_out_buffer == '') and self.connected
                        )
        
        def handle_write (self):
                "try to send a chunk of buffered output or handle error"
                buffer = self.ac_out_buffer
                sent = self.send (buffer[:self.ac_out_buffer_size])
                if sent:
                        self.ac_out_buffer = buffer[sent:]
                else:
                        self.ac_out_buffer = buffer
                                                        

class Dispatcher_with_fifo (Dispatcher):

        ac_out_buffer_size = 1 << 14 # sweet sixteen kilobytes

        def __init__ (self):
                self.ac_out_buffer = ''
                self.output_fifo = collections.deque ()
                
        def writable (self):
                "writable when there is output buffered or queued"
                return not (
                        (self.ac_out_buffer == '') and
                        not self.output_fifo and self.connected
                        )
        
        def handle_write (self):
                """pull an output fifo of single strings into the buffer, 
                send a chunk of it or close if done"""
                obs = self.ac_out_buffer_size
                buffer = self.ac_out_buffer
                fifo = self.output_fifo
                while len (buffer) < obs and fifo:
                        s = fifo.popleft ()
                        if s == None:
                                if buffer == '':
                                        self.handle_close ()
                                        return
                                
                                else:
                                        fifo.append (None)
                                        break
        
                        buffer += s
                if buffer:
                        sent = self.send (buffer[:obs])
                        if sent:
                                self.ac_out_buffer = buffer[sent:]
                        else:
                                self.ac_out_buffer = buffer
                else:
                        self.ac_out_buffer = ''
                        
        def close_when_done (self):
                "queue None if there is output queued, or handle close now"
                if self.output_fifo:
                        self.output_fifo_push (None)
                else:
                        self.handle_close ()


# Note about this implementation
#
# This is a refactored version of the asyncore's original dispatcher class,
# with a new logging facility (loginfo). The poll functions and the
# asynchronous loop have been moved to async_loop, in a single module that
# integrates a non-blocking I/O loop, a heapq scheduler loop and a loop
# through a deque of finalizations.
#
# I also :
#
# 1. added set_connection (conn, addr), moving that logic out of __init__
# 2. refactored the File_dispatcher as a set_file (fd) method
# 3. removed the now redundant set_socket (sock) method.


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/timeouts/"

import time, collections

from allegra import async_loop


class Timeouts (object):
	
	def __init__ (self, period, precision=None):
		# self.timeouts_timeout = timeout
		self.timeouts_period = max (period, async_loop.precision)
		self.timeouts_precision = precision or async_loop.precision
		self.timeouts_deque = collections.deque ()
                        
        def timeouts_push (self, reference):
                when = time.time ()
                if not self.timeouts_deque:
                        self.timeouts_start (when)
		self.timeouts_deque.append ((when, reference))
		return reference
	
        def timeouts_start (self, when):
                async_loop.schedule (
                        when + self.timeouts_precision, self.timeouts_poll
                        )
                
	def timeouts_poll (self, now):
		then = now - self.timeouts_precision - self.timeouts_period 
		while self.timeouts_deque:
			when, reference = self.timeouts_deque[0]
			if  when < then:
				self.timeouts_deque.popleft ()
				self.timeouts_timeout (reference)
			else:
				break
				
                if self.timeouts_deque:
                        return (
                                now + self.timeouts_precision, 
                                self.timeouts_poll
                                )

                self.timeouts_stop ()

	def timeouts_stop (self):
                pass # self.timeouts_timeout = None
	

# The first, simplest and probably most interesting application of Timeouts

def cached (cache, timeout, precision):
        def timedout (reference):
                try:
                        del cache[reference]
                except KeyError:
                        pass
        
        t = Timeouts (timeout, precision)
        t.timeouts_timeout = timedout
        def push (key, value):
                cache[key] = value
                t.timeouts_push (key)
                
        return push

# push, stop = timeouts.cached ({}, 60, 6)
# ...
# push (key, value)
# ...
# stop ()
        
# Note about this implementation	
#
# Time out
#
# In order to scale up and handle very large number of timeouts scheduled
# asynchronously at fixed intervals, this module provides a simple deque 
# interface that for a fifo of timeout events to poll from.
#
# Polling a timeouts queue should be scheduled recurrently at more or less
# precise intervals depending on the volume expected and the time it takes
# to handle each timeout. Your mileage may vary, but this design will scale
# up in the case of long intervals, when for each time poll only a few first 
# events at the left of the deque have timed-out.
#
# The timouts interface is applied by pns_udp to manage the 3 second timeout
# set on each statement relayed or routed by a PNS/UDP circle. There might
# be other applications, RTP protocols for instance.


# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_chat/"

import collections, socket

from allegra import async_core


def find_prefix_at_end (haystack, needle):
        "given 'haystack', see if any prefix of 'needle' is at its end."
        l = len (needle) - 1
        while l and not haystack.endswith (needle[:l]):
                l -= 1
        return l


def collect_chat (c, buffer):
        "collect a buffer for a channel or collector"
        lb = len (buffer)
        while lb:
                terminator = c.get_terminator ()
                if terminator is None or terminator == '':
                        c.collect_incoming_data (buffer)
                        buffer = ''
                elif isinstance (terminator, int):
                        if lb < terminator:
                                c.collect_incoming_data (buffer)
                                buffer = ''
                                c.set_terminator (terminator - lb)
                        else:
                                c.collect_incoming_data (buffer[:terminator])
                                buffer = buffer[terminator:]
                                c.set_terminator (0)
                                if c.found_terminator ():
                                        c.collector_stalled = True
                                        break

                else:
                        tl = len (terminator)
                        index = buffer.find (terminator)
                        if index != -1:
                                if index > 0:
                                        c.collect_incoming_data (
                                                buffer[:index]
                                                )
                                buffer = buffer[index+tl:]
                                if c.found_terminator ():
                                        c.collector_stalled = True
                                        break
                                
                        else:
                                index = find_prefix_at_end (
                                        buffer, terminator
                                        )
                                if index:
                                        if index != lb:
                                                c.collect_incoming_data (
                                                        buffer[:-index]
                                                        )
                                                buffer = buffer[-index:]
                                        break
                                        
                                else:
                                        c.collect_incoming_data (buffer)
                                        buffer = ''
                lb = len (buffer)
        return buffer


class Dispatcher (async_core.Dispatcher):

        ac_in_buffer_size = ac_out_buffer_size = 1 << 14
        
        terminator = None
        collector_stalled = False
        collector_is_simple = False
        collector_depth = 32
        
        def __init__ (self):
                self.ac_in_buffer = ''
                self.ac_out_buffer = ''
                self.output_fifo = collections.deque ()

        def __repr__ (self):
                return 'async-chat id="%x"' % id (self)
        
        def readable (self):
                "predicate for inclusion in the poll loop for input"
                return not (
                        self.collector_stalled or
                        len (self.ac_in_buffer) > self.ac_in_buffer_size
                        )

        def writable (self):
                "predicate for inclusion in the poll loop for output"
                try:
                        return not (
                                self.output_fifo[
                                        0
                                        ].producer_stalled () and
                                self.connected
                                )
                
                except:
                        return not (
                                (self.ac_out_buffer == '') and 
                                not self.output_fifo and 
                                self.connected
                                )

        def handle_read (self):
                "try to refill the input buffer and collect it"
                try:
                        data = self.recv (self.ac_in_buffer_size)
                except socket.error, why:
                        self.handle_error ()
                        return

                self.ac_in_buffer = collect_chat (
                        self, self.ac_in_buffer + data
                        )

        def handle_write (self):
                "maybe refill the output buffer and try to send it"
                obs = self.ac_out_buffer_size
                buffer = self.ac_out_buffer
                if len (buffer) < obs:
                        fifo = self.output_fifo
                        while fifo:
                                p = fifo[0]
                                if p == None:
                                        if buffer == '':
                                                fifo.popleft ()
                                                self.handle_close () 
                                                return
                                        
                                        break
                                    
                                elif type (p) == str:
                                        fifo.popleft ()
                                        buffer += p
                                        if len (buffer) < obs:
                                                continue
                                        
                                        break
                                
                                if p.producer_stalled ():
                                        break
                                
                                data = p.more ()
                                if data:
                                        buffer += data
                                        break
                                    
                                fifo.popleft ()
                if buffer:
                        sent = self.send (buffer[:obs])
                        if sent:
                                self.ac_out_buffer = buffer[sent:]
                        else:
                                self.ac_out_buffer = buffer
                else:
                        self.ac_out_buffer = ''
                        
        def close (self):
                "close the dispatcher and maybe terminate the collector"
                async_core.Dispatcher.close (self)
                if not self.collector_stalled:
                        depth = self.collector_depth
                        while depth and not self.found_terminator (): 
                                depth -= 1
                        if depth < 1:
                                self.log (
                                        '%d' % self.collector_depth,
                                        'collector-leak'
                                        )
                        
        def close_when_done (self):
                """automatically close this channel once the outgoing queue 
                is empty, or handle close now if it is allready empty"""
                if self.output_fifo:
                        self.output_fifo.append (None)
                else:
                        self.handle_close () # when done is now!

        def async_chat_push (self, p):
                "push a string or producer on the output deque"
                assert type (p) == str or hasattr (p, 'more')
                self.output_fifo.append (p)
                
        # push_with_producer = push = async_chat_push

        def async_chat_pull (self):
                "stall no more and collect the input buffer"
                self.collector_stalled = False
                if self.ac_in_buffer:
                        self.ac_in_buffer = collect_chat (
                                self, self.ac_in_buffer
                                )

        def set_terminator (self, terminator):
                "set the channel's terminator"
                self.terminator = terminator

        def get_terminator (self):
                "get the channel's terminator"
                return self.terminator

        def collect_incoming_data (self, data):
                "assert debug log of collected data"
                assert None == self.log (data, 'collect-incoming-data')

        def found_terminator (self):
                "assert debug log of terminator found"
                assert None == self.log (
                        self.get_terminator (), 'found-terminator'
                        )
                return True # do not pipeline 
        
# Note about this implementation
#
# This is a refactored version of asynchat.py as found in Python 2.4, and
# modified as to support stallable producers and collectors, loginfo and 
# finalization.
#
# Stallable Producer and Collector
#
# In order to support non-blocking asynchronous and synchronized peer, 
# the async_chat module introduces stallable collector and generalize
# the stallable producer of Medusa's proxy.
#
# Besides the fact that stallable reactors are a requirement for peers
# that do not block, they have other practical benefits. For instance,
# a channel with an collector_stalled and an empty output_fifo will not
# be polled for I/O.
#
# This implementation use collection.deque for output FIFO queues instead 
# of a class wrapper, and the push () method actually does what it is 
# supposed to do and pushes a string at the end that output queue, not a 
# Simple instance.
#
# The channel's method collect_incoming_data is called to collect data 
# between terminators. Its found_terminator method is called whenever
# the current terminator is found, and if that method returns True, then
# no more buffer will be consumed until the channel's collector_stalled
# is not set to False by a call to async_collect.



# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_client/"

import time, socket, collections

try:
        SOCKET_FAMILIES = (socket.AF_INET, socket.AF_UNIX)
except:
        SOCKET_FAMILIES = (socket.AF_INET, )

from allegra import loginfo, async_loop, async_limits
        
        
def connect (dispatcher, addr, timeout=3.0, family=socket.AF_INET):
        "create a socket, try to connect it and schedule a timeout"
        assert (
                not dispatcher.connected and timeout > 0 and
                family in SOCKET_FAMILIES
                )
        dispatcher.client_when = time.time ()
        dispatcher.client_timeout = timeout
        try:
                dispatcher.create_socket (family, socket.SOCK_STREAM)
                dispatcher.connect (addr)
        except:
                dispatcher.loginfo_traceback ()
                dispatcher.handle_error ()
                return False
                
        assert None == dispatcher.log ('connect', 'debug')
        def connect_timeout (when):
                "if not connected and not closing yet, handle close"
                if not dispatcher.connected and not dispatcher.closing:
                        assert None == dispatcher.log (
                                'connect-timeout %f seconds' % (
                                        when - dispatcher.client_when
                                        ), 'debug'
                                )
                        dispatcher.handle_close ()

        async_loop.schedule (
                dispatcher.client_when + timeout, connect_timeout
                )
        return True


def reconnect (dispatcher):
        if dispatcher.addr:
                dispatcher.closing = False
                return connect (
                        dispatcher, 
                        dispatcher.addr, 
                        dispatcher.client_timeout, 
                        dispatcher.family_and_type[0]
                        )
                        
        return False
                

class Connections (loginfo.Loginfo):
        
        "a connection manager for async_client.Dispatcher instances"

        ac_in_meter = ac_out_meter = 0
        client_errors = client_when = client_dispatched = 0
        
        def __init__ (
                self, timeout=3.0, precision=1.0, family=socket.AF_INET
                ):
                "initialize a new client manager"
                assert (
                        timeout > 0 and precision > 0 and
                        family in SOCKET_FAMILIES
                        )
                self.client_managed = {}
                self.client_timeout = timeout
                self.client_precision = precision
                self.client_family = family
                resolved (self)
                inactive (self, timeout)
                
        def __call__ (self, dispatcher, name):
                "registed, decorate and connect a new dispatcher"
                if self.client_connect (dispatcher, name):
                        now = time.time ()
                        dispatcher.async_client = self
                        self.client_decorate (dispatcher, now)
                        key = id (dispatcher)
                        self.client_managed[key] = dispatcher
                        dispatcher.client_key = key
                        if len (self.client_managed) == 1:
                                self.client_start (now)
                else:
                        self.client_errors += 1
                return dispatcher
                
        def client_connect (self, dispatcher, name):
                "resolve and/or connect a dispatcher"
                dispatcher.client_name = name
                addr = self.client_resolved (name)
                if addr != None:
                        return connect (
                                dispatcher, addr, 
                                self.client_timeout, self.client_family
                                )

                if self.client_resolve == None:
                        self.client_unresolved (dispatcher, addr)
                        return False
                
                def resolve (addr):
                        if addr == None:
                                self.client_unresolved (dispatcher, name)
                                return
                        
                        if not connect (
                                dispatcher, addr, 
                                self.client_timeout, self.client_family
                                ):
                                self.client_errors += 1
                                
                self.client_resolve (name, resolve)
                return True
        
        def client_reconnect (self, dispatcher):
                dispatcher.closing = False
                self (dispatcher, dispatcher.client_name)
                return dispatcher.closing
        
        def client_unresolved (self, dispatcher, name):
                "assert debug log and close an unresolved dispatcher"
                assert None == dispatcher.log (
                        '%r unresolved' % (name, ), 'debug'
                        )
                self.client_errors += 1
                dispatcher.handle_close ()

        def client_start (self, when):
                "handle the client management startup"
                self.client_when = when
                async_loop.schedule (
                        when + self.client_precision, self.client_manage
                        )
                assert None == self.log ('start', 'debug')

        def client_manage (self, when):
                "test limits overflow, recure or stop"
                for dispatcher in self.client_dispatchers ():
                        if self.client_limit (dispatcher, when):
                                self.client_overflow (dispatcher)
                if self.client_managed:
                        return (
                                when + self.client_precision,
                                self.client_manage
                                ) # continue to defer
                
                self.client_stop (when)
                return None
        
        def client_dispatchers (self):
                "return a list of managed dispatchers"
                return self.client_managed.values ()
                        
        def client_overflow (self, dispatcher):
                "assert debug log and close an overflowed dispatcher"
                assert None == dispatcher.log ('limit overflow', 'debug')
                dispatcher.handle_close ()
                
        def client_meter (self, dispatcher):
                "assert debug log and account I/O meters of a dispatcher"
                assert None == dispatcher.log (
                        'in="%d" out="%d"' % (
                                dispatcher.ac_in_meter, 
                                dispatcher.ac_out_meter
                                ),  'debug'
                        )
                self.ac_in_meter += dispatcher.ac_in_meter
                self.ac_out_meter += dispatcher.ac_out_meter
                self.client_dispatched += 1

        def client_close (self, dispatcher):
                "remove the dispatcher from cache and meter dispatched"
                del self.client_managed[dispatcher.client_key]
                self.client_meter (dispatcher)
                dispatcher.async_client = None

        def client_stop (self, when):
                "handle the client management stop"
                assert None == self.log (
                        'stop errors="%d" dispatched="%d"'
                        ' seconds="%f" in="%d" out="%d"' % (
                                self.client_errors,
                                self.client_dispatched,
                                (when - self.client_when),
                                self.ac_in_meter,
                                self.ac_out_meter
                                ), 'debug')
                self.client_errors = self.client_dispatched = \
                        self.ac_in_meter = self.ac_out_meter = 0

        def close_when_done (self):
                "close all client dispatchers when done"
                for dispatcher in self.client_dispatchers ():
                        dispatcher.close_when_done ()
                        

class Cache (Connections):

        "a cache of managed connections"

        def __init__ (
                self, timeout=3.0, precision=1.0, family=socket.AF_INET
                ):
                "initialize a new client cache"
                assert (
                        timeout > 0 and precision > 0 and
                        family in SOCKET_FAMILIES
                        )
                self.client_managed = {}
                self.client_timeout = timeout
                self.client_precision = precision
                self.client_family = family
                resolved (self)
                inactive (self, timeout)
                
        def __call__ (self, Dispatcher, name):
                """return a cached or a new dispatcher, maybe resolving and
                connecting it first, closing it on connection error or if
                it's socket address cannot be resolved"""
                try:
                        return self.client_managed[name]
                        
                except KeyError:
                        pass
                dispatcher = Dispatcher ()
                if self.client_connect (dispatcher, name):
                        now = time.time ()
                        dispatcher.async_client = self
                        self.client_decorate (dispatcher, now)
                        self.client_managed[name] = dispatcher
                        dispatcher.client_key = name
                        if len (self.client_managed) == 1:
                                self.client_start (now)
                else:
                        self.client_errors += 1
                return dispatcher


class Pool (Connections):
        
        "a pool of managed connections"
        
        def __init__ (
                self, Dispatcher, name, 
                size=2, timeout=3.0, precision=1.0, family=socket.AF_INET
                ):
                "initialize a new client pool"
                assert (
                        type (size) == int and size > 1 and
                        timeout > 0 and precision > 0 and
                        family in SOCKET_FAMILIES
                        )
                self.client_managed = []
                self.client_pool = size
                self.client_name = name
                self.client_called = 0
                self.Client_dispatcher = Dispatcher
                self.client_timeout = timeout
                self.client_precision = precision
                self.client_family = family
                resolved (self)
                inactive (self, timeout)
                
        def __call__ (self):
                """return the next dispatcher pooled or instanciate a new
                one, maybe resolving and connecting it first, closing it on 
                connection error or if it's socket address cannot be 
                resolved"""
                size = len (self.client_managed)
                if size >= self.client_pool:
                        self.client_called += 1
                        return self.client_managed[self.client_called % size]
                
                now = time.time ()
                dispatcher = self.Client_dispatcher ()
                if self.client_connect (dispatcher, self.client_name):
                        dispatcher.async_client = self
                        self.client_decorate (dispatcher, now)
                        self.client_managed.append (dispatcher)
                        if len (self.client_managed) == 1:
                                self.client_start (now)
                else:
                        self.client_errors += 1
                return dispatcher
        
        def client_reconnect (self, dispatcher):
                return False # useless for a cached dispatcher!

        def client_dispatchers (self):
                "return a list of dispatchers pooled"
                return list (self.client_managed)
                        
        def client_close (self, dispatcher):
                "remove the dispatcher from pool and increment dispatched"
                self.client_meter (dispatcher)
                self.client_managed.remove (dispatcher)
                dispatcher.async_client = None


def resolved (connections):
        "allways resolved for unresolved dispatcher address"
        connections.client_resolved = (lambda addr: addr)
        connections.client_resolve = None
        return connections


def meter (dispatcher, when):
        "decorate a client dispatcher with stream meters"
        async_limits.meter_recv (dispatcher, when)
        async_limits.meter_send (dispatcher, when)
        def close ():
                del (
                        dispatcher.recv, 
                        dispatcher.send, 
                        dispatcher.close
                        )
                dispatcher.close ()
                dispatcher.async_client.client_close (dispatcher)
                
        dispatcher.close = close
        
def no_limit (dispatcher, when):
        return False
        
        
def unlimited (connections):
        "meter I/O for unlimited client streams"
        connections.client_decorate = meter
        connections.client_limit = no_limit
        return connections
        

def inactive (connections, timeout):
        "meter I/O and limit inactivity for client streams"
        assert timeout > 0
        def decorate (dispatcher, when):
                meter (dispatcher, when)
                dispatcher.limit_inactive = connections.client_inactive
                
        connections.client_decorate = decorate
        connections.client_inactive = timeout
        connections.client_limit = async_limits.inactive
        return connections


def limited (connections, timeout, inBps, outBps):
        "throttle I/O and limit inactivity for managed client streams"
        assert (
                timeout > 0 and
                type (inBps ()) == int and inBps () > 0 and
                type (outBps ()) == int and outBps () > 0
                )
        def throttle (dispatcher, when):
                "decorate a client dispatcher with stream limits"
                async_limits.meter_recv (dispatcher, when)
                async_limits.meter_send (dispatcher, when)
                dispatcher.limit_inactive = timeout
                async_limits.throttle_readable (
                        dispatcher, when, connections.ac_in_throttle_Bps
                        )
                async_limits.throttle_writable (
                        dispatcher, when, connections.ac_out_throttle_Bps
                        )
                def close ():
                        del (
                                dispatcher.recv, 
                                dispatcher.send, 
                                dispatcher.readable,
                                dispatcher.writable,
                                dispatcher.close
                                )
                        dispatcher.close ()
                        dispatcher.async_client.client_close (dispatcher)
                        
                dispatcher.close = close

        connections.client_decorate = throttle
        connections.ac_in_throttle_Bps = inBps
        connections.ac_out_throttle_Bps = outBps
        connections.client_limit = async_limits.limit
        return connections


def rationed (connections, timeout, inBps, outBps):
        "ration I/O and limit inactivity for managed client streams"
        assert (
                timeout > 0 and
                type (inBps) == int and inBps > 0 and
                type (outBps) == int and outBps > 0
                )
        connections.ac_in_ration_Bps = inBps
        connections.ac_out_ration_Bps = outBps
        def throttle_in ():
                return int (connections.ac_in_ration_Bps / max (len (
                        connections.client_managed
                        ), 1))

        def throttle_out ():
                return int (connections.ac_out_ration_Bps / max (len (
                        connections.client_managed
                        ), 1))

        return limited (connections, timeout, throttle_in, throttle_out)
        


class Pipeline (object):
        
        "a pipeline mix-in for dispatcher"

        #pipeline_sleeping = False
        pipeline_pipelining = False
        pipeline_keep_alive = False

        def pipeline_set (self, requests=None, responses=None):
                "set new requests and responses deque"
                self.pipeline_requests = requests or collections.deque ()
                self.pipeline_responses = responses or collections.deque ()

        #def pipeline (self, request):
        #        "pipeline a new request, wake up if sleeping"
        #        self.pipeline_requests.append (request)
        #        if self.pipeline_sleeping:
        #                self.pipeline_sleeping = False
        #                self.pipeline_wake_up ()

        def pipeline_wake_up (self):
                requests = self.pipeline_requests
                if self.pipeline_pipelining and len (requests) > 1:
                        self.pipeline_requests = deque ()
                        self.output_fifo.extend ((
                                request[0] for request in requests
                                ))
                        self.pipeline_responses.extend (requests)
                else:
                        request = self.pipeline_requests.popleft ()
                        self.output_fifo.append (request[0])
                        self.pipeline_responses.append (request)
                
                        


# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_loop/"

import gc, select, errno, time, collections, heapq

from allegra import loginfo


Exit = KeyboardInterrupt


# Poll I/O

def _io_select (map, timeout, limit):
        "poll for I/O a limited number of writable/readable dispatchers"
        r = []
        w = []
        concurrent = map.items ()
        rest = limit - len (concurrent)
        if rest < 0:
                concurrent = concurrent[:limit]
        else:
                rest = 0
        for fd, dispatcher in concurrent:
                if dispatcher.readable ():
                        r.append (fd)
                if dispatcher.writable ():
                        w.append (fd)
        if len (r) + len (w) == 0:
                time.sleep (timeout)
                return limit - rest, 0

        try:
                r, w, e = select.select (r, w, [], timeout)
        except select.error, err:
                if err[0] != errno.EINTR:
                    raise
                    
                else:
                    return limit - rest, 0

        for fd in r:
                try:
                        dispatcher = map[fd]
                except KeyError:
                        continue

                try:
                        dispatcher.handle_read_event ()
                except Exit:
                        raise
                        
                except:
                        dispatcher.handle_error ()
        for fd in w:
                try:
                        dispatcher = map[fd]
                except KeyError:
                        continue

                try:
                        dispatcher.handle_write_event ()
                except Exit:
                        raise 
                        
                except:
                        dispatcher.handle_error ()
        return limit - rest, len (r) + len (w)
        #
        # note that the number of distinct active dispatchers may actually
        # be lower than the one reported: to get an exact cound would
        # require to use sets, something like: len (set (r) + set (w))

def _io_poll (map, timeout, limit):
        "poll for I/O a limited number of writable/readable dispatchers"
        timeout = int (timeout*1000)
        pollster = select.poll ()
        R = select.POLLIN | select.POLLPRI
        W = select.POLLOUT
        RW = R | W
        concurrent = map.items ()
        rest = limit - len (concurrent)
        if rest < 0:
                concurrent = concurrent[:limit]
        else:
                rest = 0
        for fd, dispatcher in concurrent:
                if dispatcher.readable ():
                        if dispatcher.writable ():
                                pollster.register (fd, RW)
                        else:
                                pollster.register (fd, R)
                elif dispatcher.writable ():
                        pollster.register (fd, W)
        try:
                p = pollster.poll (timeout)
        except select.error, err:
                if err[0] != errno.EINTR:
                        raise
                        
        else:
                for fd, flags in p:
                        try:
                                dispatcher = map[fd]
                        except KeyError:
                                continue
        
                        try:
                                if flags & R:
                                        dispatcher.handle_read_event ()
                                if flags & W:
                                        dispatcher.handle_write_event ()
                        except Exit:
                                raise 
                                
                        except:
                                dispatcher.handle_error()
        return limit - rest, len (p)


# select the best I/O poll function available for this system

if hasattr (select, 'poll'):
	_io = _io_poll
else:
	_io = _io_select


# Poll Memory (Finalizations, ie: CPython __del__ decoupled)

_finalized = collections.deque ()

def _finalize ():
	"call all finalizations queued"
        while True:
                try:
                        finalized = _finalized.popleft ()
                except IndexError:
                        break
                       
                try:
                        finalized.finalization = finalized.finalization (
                               finalized
                               ) # finalize and maybe continue ...
                except Exit:
                        finalized.finalization = None
                        raise
                
                except:
                        finalized.finalization = None
                        loginfo.traceback () # log exception
        
	
# Poll Time (Scheduled Events)

precision = 0.1

_scheduled = []

def _clock ():
	"call all events scheduled before now, maybe recurr in the future"
	now = time.time ()
        future = now + precision
	while _scheduled:
		# get the next defered ...
		event = heapq.heappop (_scheduled)
		if event[0] > now:
			heapq.heappush (_scheduled, event)
			break  # ... nothing to defer now.

		try:
			# ... do defer and ...
			continued = event[1] (event[0])
                except Exit:
                        raise
                
		except:
			loginfo.traceback ()
		else:
			if continued != None:
                                # ... maybe recurr in the future
                                if continued[0] < future:
                                        continued = (future, continued[1])
				heapq.heappush (_scheduled, continued) 


# Poll Signals (Exceptions Handler)

_catchers = []

def _catched ():
        "call async_loop.Exit exception catchers"
        assert None == loginfo.log ('async_catch', 'debug')
        if _catchers:
                for catcher in tuple (_catchers):
                        if catcher ():
                                _catchers.remove (catcher)
                return True
        
        if __debug__:
                for dispatcher in _dispatched.values ():
                        loginfo.log (
                                '%r' % dispatcher, 'undispatched'
                                )
                for event in _scheduled:
                        loginfo.log (
                                '%r' % (event, ), 'unscheduled'
                                )
                for finalized in _finalized:
                        loginfo.log (
                                '%r' % (finalized, ), 'unfinalized'
                                )
        return False


# Application Programming Interfaces

def schedule (when, scheduled):
        "schedule a call to scheduled after when"
        heapq.heappush (_scheduled, (when, scheduled))        
        

def catch (catcher):
        "register an catcher for the Exit exception"
        _catchers.append (catcher)


concurrency = 512

_dispatched = {}

def dispatch ():
        "dispatch I/O, time and finalization events"
        assert None == loginfo.log ('async_dispatch_start', 'debug')
        while _dispatched or _scheduled or _finalized or gc.collect () > 0:
                try:
                        _io (_dispatched, precision, concurrency)
                        _clock ()
                        _finalize ()
                except Exit:
                        if not _catched ():
                                break
                
                except:
                        loginfo.traceback ()
        
        assert None == loginfo.log ('async_dispatch_stop', 'debug')
   
   
def io_meter (loop, when=None):
        "decorate the loop module's I/O poll function with meters, log info"
        loop._io_when = when or time.time ()
        loop._io_run = 0
        loop._io_load = loop._io_concurrency = loop._io_activity = 0.0
        def _io_metering (map, timeout, limit):
                loop._io_run += 1
                dispatched = len (loop._dispatched)
                concurrent, polled = _io (map, timeout, limit)
                if concurrent > 0:
                        loop._io_activity += float (active) / concurrent
                        concurrent = float (concurrent)
                        loop._io_load += concurrent / limit
                        loop._io_concurrency += concurrent / dispatched

        loop._io_metered = loop._io
        loop._io = _io_metering
        loginfo.log ('io-metered', 'info')
        
def io_meters (loop):
        "return statistics about a metered I/O loop"
        return (
                loop._io_when, loop._io_run,
                loop._io_run / (time.time () - loop._io_when),
                loop._io_load / loop._io_run,
                loop._io_concurrency / loop._io_run,
                loop._io_activity / loop._io_run
                )

def io_unmeter (loop):
        "log meter statistics and remove the metering decoration"
        meters = io_meters (loop)
        loginfo.log (
                'io-unmetered'
                ' seconds="%f" run="%d"'
                ' rps="%f" load="%f"'
                ' concurrency="%f" activity="%f"' % meters, 'info'
                )
        del (
                loop._io_when, loop._io_run, 
                loop._io_load, _io_concurrency, loop._io_activity
                )
        loop._io = loop._io_metered
        return meters


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/loginfo/"

import sys

from allegra import netstring, prompt


def _write_maybe_flush (file):
        assert hasattr (file, 'write')
	if hasattr (file, 'flush'):
                def write (data):
                        file.write (data)
                        file.flush ()
                        return 
                
		return write

	return file.write


class Logger (object):

	"Loginfo's log dispatcher implementation"
	
	def __init__ (self):
                self.loginfo_stdout = _write_maybe_flush (sys.stdout)
                self.loginfo_stderr = _write_maybe_flush (sys.stderr)
		self.loginfo_categories = {}

	def loginfo_netstrings (self, data, info=None):
		"log netstrings to STDOUT, a category handler or STDERR"
                assert type (data) == str
		if info == None:
			self.loginfo_stdout ('%d:%s,' % (len (data), data))
		elif self.loginfo_categories.has_key (info):
			self.loginfo_categories[info] (
				'%d:%s,' % (len (data), data)
				)
		else:
                        assert type (info) == str
			encoded = netstring.encode ((info, data))
			self.loginfo_stderr (
				'%d:%s,' % (len (encoded), encoded)
				)

	def loginfo_netlines (self, data, info=None):
		"log netoutlines to STDOUT, a category handler or STDERR"
                assert type (data) == str
		if info == None:
			self.loginfo_stdout (netstring.netlines (data))
		elif self.loginfo_categories.has_key (info):
			self.loginfo_categories[info] (
				netstring.netlines (data)
				)
		else:
			assert type (info) == str
			self.loginfo_stderr (netstring.netlines (
				netstring.encode ((info, data))
				))

	if __debug__:
		log = loginfo_netlines
	else:
		log = loginfo_netstrings


# the facility instance and module interfaces

logger = Logger ()

def log (*args):
        logger.log (*args)


def compact_traceback (ctb):
	"encode a compact traceback tuple as netstrings"
 	return netstring.encode ((
                ctb[0], 
 		netstring.encode ([' | '.join (x) for x in ctb[2]]),
                ctb[1]
 		)), 'traceback'
                 
def classic_traceback (ctb=None):
        return netstring.encode ((
                netstring.encode ([
                        'File "%s", line %s, in %s' % (
                                tb[0] or '<string>', tb[2], tb[1]
                                )
                        for tb in ctb[2]
                        ]), '%s: %s' % ctb[:2]
                )), 'Traceback (most recent call last):'

traceback_encode = compact_traceback
        
def traceback (ctb=None):
        "return a compact traceback and log it in the 'traceback' category"
        if ctb == None:
                ctb = prompt.compact_traceback ()
        logger.log (*traceback_encode (ctb))
        return ctb


# redirection of all Python standard outputs to the logger

class _STDOUT (object):

        def write (self, line):
                logger.log (line)
        
sys.stdout = _STDOUT ()

def _displayhook (value):
        if value != None:
                logger.log ('%r' % (value,))
        
sys.displayhook = _displayhook
        
class _STDERR (object):

        def write (self, line):
                logger.log (line, 'stderr')
        
def _excepthook (*exc_info):
        traceback (prompt.compact_traceback (exc_info))

sys.excepthook = _excepthook


# a class interface to mix in

class Loginfo (object):

	loginfo_logger = logger
	
	def __repr__ (self):
		return '%s id="%x"' % (
			self.__class__.__name__, id (self)
			)

	def loginfo_log (self, data, info=None):
		"""log a message with this instance's __repr__ and an 
		optional category"""
		self.loginfo_logger.log (netstring.encode ((
			'%r' % self, '%s' % data
			)), info)

	log = loginfo_log

	def loginfo_null (self, data, info=None):
                "drop the message to log"
                pass
                
        def loginfo_logging (self):
        	"return True if the instance is logging"
        	return self.log != self.loginfo_null

	def loginfo_toggle (self, logging=None):
		"toggle logging on/off for this instance"
		if logging == None:
			if self.log == self.loginfo_null:
				try:
					del self.log
				except:
					self.log = self.loginfo_log
			else:
				try:
					del self.log
				except:
					self.log = self.loginfo_null
			return self.log != self.loginfo_null
			
		if logging == True and self.log == self.loginfo_null:
			self.log = self.loginfo_log
		elif logging == False and self.log == self.loginfo_log:
			self.log = self.loginfo_null
		return logging

	def loginfo_traceback (self, ctb=None):
		"""return a compact traceback tuple and log it encoded as 
		netstrings, along with this instance's __repr__, in the
		'traceback' category"""
		if ctb == None:
			ctb = prompt.compact_traceback ()
		self.loginfo_log (*traceback_encode (ctb))
		return ctb
                

def toggle (logging=None, Class=Loginfo):
	"toggle logging on/off for the Class specified or Loginfo"
	if logging == None:
		if Class.log == Class.loginfo_null:
			Class.log = Class.loginfo_log
			return True
			
		Class.log = Class.loginfo_null
		return False
		
	if logging == True and Class.log == Class.loginfo_null:
		Class.log = Class.loginfo_log
	elif logging == False and Class.log == Class.loginfo_log:
		Class.log = Class.loginfo_null
	return logging
        

# SYNOPSIS
#
# >>> from allegra import loginfo
#>>> loginfo.log ('message')
# message
# >>> loginfo.log ('message', 'info')
# info
# message
#	
# >>> try:
# ...    foobar ()
# ... except:
# ...    ctb = loginfo.traceback ()
# traceback
#   exceptions.NameError
#   name 'foobar' is not defined
#     <stdin> | ? | 2
#
# >>> logged = loginfo.Loginfo ()
# >>> logged.log ('message')
# Loginfo id="8da4e0"
# message
#	
# >>> logged.log ('message', 'info')
# info
#   Loginfo id="8da4e0"
#   message
#
# The Loginfo interface and implementation provide a simpler, yet more
# powerfull and practical logging facility than the one currently integrated
# with Python.
#
# First is uses netstrings instead of CR or CRLF delimeted lines for I/O
# encoding, solving ahead many problems of the log consumers. Second it
# is well adapted to a practical development cycle and by defaults implement
# a simple scheme that suites well a shell pipe like:
#
#	pipe < input 1> output 2> errors
#
# However "slow" this Python facily is, it offers a next-best trade-off
# between performance in production and flexibility for both debugging and
# maintenance. All debug and information logging can be effectively disabled,
# either skipped in case of:
#
#	assert None == loginfo.log (...)
#
# or simply dropped, yielding not access to possibly blocking or buffering
# process handling the log output. The ability to filter out or filter in
# log at the instance or class level is enough to satisfy the most demanding
# application administrator (as for categories, there are none defined, do
# as it please you ;-).
#
# Last but not least, the loginfo_log is compatible with asyncore logging 
# interfaces (which is no wonder, it is inspired from Medusa's original
# logging facility module).
#
#
# CAVEAT!
#
# since Loginfo instances update their own namespace with bound methods,
# they actually cycle, referencing themselves. so, when toggling off is
# a requirement if you need to finalize an instance that has been 
# explicitely logged in or out.
#
# the trade-off is the following: most toggling happens at the class level
# in practice. From development to production, instance loginfo toggling
# will disapear. And if you ask to log an object at runtime, it's a practical
# way to *not* finalize it: you're observing something! And that's even more
# true if you have to manipulate instances in production ...
#
# Simple but not obvious :-)
#
#
# One Logger Only
#
# As time and applications go by, it made sense to do with only one logger
# facility, a root logger. That's the model of log4j, but with an extra
# simplification: STDOUT and STDERR are replaced by loginfo's interfaces
# and implementation.
#
# If you 
#
# >>> print "something"
#
# in your application, it will actually be logged without categories nor
# context, directly to STDOUT. In the debug outline mode, this is exactly
# similar to writing "something" out with a newline at the end. However,
# in optimized mode, those print statement will be encoded in netstrings
# and without newline.
#
# Also, when an uncatched exception is raised, it will be logged as an
# outline or a netstring (or a multiline classic traceback, for people
# who really need that at debug time).
#
# The rational is that if your application needs to write to a file in all
# case, it either does not need a log or that file should not be STDOUT.
# Similarly, if your application does demand an error log, there is no
# reason not to have them writen to STDERR, preferrably in a format that is
# simple and safe to parse and present.
#
# 
# Practicality beats purity
#
# Satisfying a demand of Andrew Dalke, loginfo gives the option of classic
# and compact tracebacks. The later are encoded in netstring and outlined
# at debug time, the former are allways multiline strings (the original
# format) that can be parsed by existing Python development tools.
#
# If you do use those tools to lookup source code at development time, do
#
# if __debug__:
#        Loginfo.loginfo_traceback = Loginfo.loginfo_classic_traceback
#        loginfo.traceback = loginfo._classic_traceback
#
# 


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/allegra/doc/"

__all__ = [
        #
        # Asynchronous Network Peer Programming
        #
        'netstring', 'prompt', 'loginfo', 
        'async_loop', 'finalization', 'async_core',
        'select_trigger', 'thread_loop', 'sync_stdio', 
        'async_net', 'async_chat', 'producer', 'collector', 'reactor', 
        'async_limits', 'async_server', 'async_client', 
        'synchronized', 'timeouts',
        #
        # Major Internet Application Protocols
        #
        'ip_peer', 'dns_client', 
        'tcp_server', 'tcp_client', 
        'mime_headers', 'mime_reactor',
        # 'smtp_client', 'pop_client', 'nnrp_client', 
        'http_reactor', 'http_client', 'http_server', 
        'xml_dom', 'xml_unicode', 'xml_utf8', 'xml_reactor',
        #
        # PNS, The Reference Implementation
        #
        'sat', 'pns_model', 
        'pns_sat', 'pns_mime', 'pns_xml', 'pns_rss', # 'pns_html', 
        'pns_tcp', 'pns_resolution', 'pns_inference', 'pns_udp', 
        'pns_peer', 'pns_client', 'pns_articulator', 
        #
        # Allegra PRESTo!
        #
        'presto', 'presto_http', 
        'presto_prompt', 'presto_pns', # 'presto_bsddb', 
        #
        # The Last DNS Application
        #
        # 'pns_dns'
        ]

__author__ = 'Laurent A.V. Szyster <contact@laurentszyster.be>'


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/collector/"


from allegra import async_chat, loginfo


class Null (object):

        # collect data to /dev/null

        collector_is_simple = True

        def collect_incoming_data (self, data):
                return

        def found_terminator (self):
                return True


DEVNULL = Null ()


class Loginfo (object):
        
        # collect data to loginfo
        
        collector_is_simple = True

        def __init__ (self, info=None):
                self.info = info
        
        def collect_incoming_data (self, data):
                loginfo.log (data, self.info)
                
        def found_terminator (self):
                return True # final!


LOGINFO = Loginfo ()
                

class File (object):
        
        collector_is_simple = True
        
        def __init__ (self, file):
                self.file = file
                self.collect_incoming_data = self.file.write

        def found_terminator (self):
                self.file.close ()
                self.collect_incoming_data = None
                return True
                
                
def devnull (data): pass
                
class Limited (object):
        
        collector_is_simple = True
        
        def __init__ (self, limit):
                self.data = ''
                self.limit = limit

        def collect_incoming_data (self, data):
                self.limit -= len (data)
                if self.limit > 0:
                        self.data += data
                else:
                        self.collect_incoming_data = devnull
                        # 
                        # just don't do anything after the limit, not even
                        # testing for it ;-)

        def found_terminator (self):
                return True
                

class Codec_decoder (object):
        
        # Decode collected data using the codecs' decode interface:
        #
        #        import codecs
        #        Codec_decoder (collector, codecs.lookup ('zlib')[1])
        #
        # Note that the decode function *must* decode byte strings, not
        # UNICODE strings.
        
        collector_is_simple = True
        
        def __init__ (self, collector, decode):
                assert collector.collector_is_simple
                self.collector = collector
                self.decode = decode
                self.buffer = ''
        
        def collect_incoming_data (self, data):
                if self.buffer:
                        decoded, consumed = self.decode (self.buffer + data)
                        consumed -= len (self.buffer)
                else:
                        decoded, consumed = self.decode (data)
                self.collector.collect_incoming_data (decoded)
                if consumed < len (data) + 1:
                        self.buffer = data[consumed:]
                        
        def found_terminator (self, data):
                if self.buffer:
                        decoded, consumed = self.decode (self.buffer)
                        if decoded:
                                self.collector.collect_incoming_data (decoded)
                return self.collector.found_terminator ()
       

class Padded_decoder (object):
        
        # Collect padded blocks to decode, for instance:
        #
        #        import base64
        #        Padded_decoder (collector, 20, base64.b64decode)
        #
        # because padding does matter to the base binascii implementation,
        # and is not handled by the codecs module, a shame when a large
        # XML string is encoded in base64 and should be decoded and parsed
        # asynchronously. Padding is also probably a requirement from block
        # cypher protocols and the likes.
        
        collector_is_simple = True
        
        def __init__ (self, collector, padding, decode):
                assert collector.collector_is_simple
                self.collector = collector
                self.padding = padding
                self.decode = decode
                self.buffer = ''
        
        def collect_incoming_data (self, data):
                lb = len (self.buffer) + len (data) 
                if lb < self.padding:
                        self.buffer += data
                        return

                tail = lb % self.padding
                if self.buffer:
                        if tail:
                                self.buffer = data[-tail:]
                                self.collector.collect_incoming_data (
                                        self.decode (
                                                self.buffer + data[:-tail]
                                                )
                                        )
                        else:
                                self.collector.collect_incoming_data (
                                        self.decode (self.buffer + data)
                                        )
                elif tail:
                        self.buffer = data[-tail:]
                        self.collector.collect_incoming_data (
                                self.decode (data[:-tail])
                                )
                else:
                        self.collector.collect_incoming_data (
                                self.decode (data)
                                )
        
        def found_terminator (self):
                if self.buffer:
                        self.collector.collect_incoming_data (
                                self.decode (self.buffer)
                                )
                        self.buffer = ''
                return self.collector.found_terminator ()
        
        
class Simple (object):

        collector_is_simple = True
        terminator = None
        buffer = ''

        def get_terminator (self):
                return self.terminator

        def set_terminator (self, terminator):
                self.terminator = terminator

        def collect_incoming_data (self, data):
                self.buffer = async_chat.collect_chat (
                        self.collector, self.buffer + data
                        )
                        
        def found_terminator (self):
                if self.buffer:
                        async_chat.collect_chat (self.collector, self.buffer)
                return True # allways final


def bind_simple (cin, cout):
        "bind to a simple collector until found_terminator is True"
        def found_terminator ():
                if cout.found_terminator ():
                        del (
                                cin.collect_incoming_data,
                                cin.found_terminator
                                )
                        return cin.found_terminator ()
                
                return False

        cin.collect_incoming_data = cout.collect_incoming_data
        cin.found_terminator = found_terminator
        return cin

def bind_complex (cin, cout):
        "bind to a complex collector until found_terminator is True"
        cout.set_terminator = cin.set_terminator
        cout.get_terminator = cin.get_terminator
        cout.collector = cin
        cin.set_terminator (cout.get_terminator ())
        cin.collect_incoming_data = cout.collect_incoming_data
        def found_terminator ():
                if cout.found_terminator ():
                        del (
                                cout.set_terminator,
                                cout.get_terminator,
                                cout.collector,
                                cin.collect_incoming_data,
                                cin.found_terminator
                                )
                        return cin.found_terminator ()
                
                return False
        
        cin.found_terminator = found_terminator
        return cin

def bind (cin, cout):
        "bind a complex to a collector until found_terminator is True"
        assert not cin.collector_is_simple
        if cout.collector_is_simple:
                return bind_simple (cin, cout)

        return bind_complex (cin, cout)


def simplify (cin, cout):
        couple = Simple ()
        bind_complex (couple, cout)
        return bind_simple (cin, couple)

def simple (collected):
        return bind_simple (Simple (), collected)        


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/finalization/"


from allegra import loginfo, async_loop


# Finalize

class Finalization (object):

        finalization = None
        
        async_finalized = async_loop._finalized

        def __del__ (self):
                if self.finalization != None:
                         self.async_finalized.append (self)


def collect ():
        import gc
        collected = gc.collect ()
        if collected == 0:
                return
        
        assert None == loginfo.log ('%d' % collected, 'collected')
        for cycle in gc.garbage:
                try:
                        cylce.finalization = None
                except:
                        pass
                assert None == loginfo.log ('%r' % cycle, 'garbage')
        

# Branch

class Branch (Finalization):

	def __init__ (self, finalizations):
		self.finalizations = finalizations

	def __call__ (self, finalized):
		for finalization in self.finalizations:
			finalization (finalized)

def branch (branched, finalization):
	try:
		branched.finalization.finalizations.append (finalization)
	except:
		branched.finalization = Branch ([
			branched.finalization, finalization
			])
		
		
# Continue

class Continuation (object):

        finalization = None
        async_finalized = async_loop._finalized
        
        def __call__ (self): pass

        def __del__ (self):
                if self.finalization != None:
                         self.async_finalized.append (self)

	
def continuation (finalizations):
	"combines continuations into one execution path"
        i = iter (finalizations)
	first = continued = i.next ()
        try:
        	while True:
                        continued.finalization = i.next ()
                        continued = continued.finalization
        except StopIteration:
                pass
	return first, continued


class Continue (Continuation):
        
	def __init__ (self, finalizations):
		self.__call__, self.continued = continuation (
                        finalizations
                        )

# Join
#
# the equivalent of "thread joining" with finalization does not really need 
# a specific interface because it is enough to set the "joining" finalized
# as the finalization of all "joined" finalized.
#
#        joined.finalization = joining
#
# how simple ...

def join (finalizations, continuation):
        def finalize (finalized):
                for joined in finalizations:
                        joined.finalization = continuation
        return finalize
                

class Join (Continuation):

        def __init__ (self, finalizations):
                self.finalizations = finalizations
        
        def __call__ (self, finalized):
                if self.finalizations:
                        # start
                        for joined in self.finalizations:
                                joined.finalization = self
                        self.finalizations = None
                #
                # join
                


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/prompt/"

import sys, types


def compact_traceback (exc_info=None):
        """return a compact traceback tuple from sys.exc_info(), like:
        
        (['error name',
                ('filename', 'lineno', 'function'),
                ...
                ], 'error message')
                
        a compact traceback is a simple data structure made of 8-bit byte 
        strings, ready to be serialized."""
        t, v, tb = exc_info or sys.exc_info ()
        if type (t) == types.ClassType:
                t = t.__name__
        elif type (t) != str:
                t = str (t)
        tbinfo = []
        assert tb # Must have a traceback ?
        while tb:
                tbinfo.append ((
                        tb.tb_frame.f_code.co_filename,
                        tb.tb_frame.f_code.co_name,
                        str (tb.tb_lineno)
                        ))
                tb = tb.tb_next
        del tb # just to be safe ?
        return t, str (v), tbinfo


def python_eval (co, env):
        """try to eval the compiled co in the environement env
        return either ('eval', result) or ('excp', traceback)"""
        try:
                return ('eval', eval (co, env))

        except:
                return ('excp', compact_traceback ())


def python_exec (co, env):
        """try to exec the compiled co in the environement env
        return either ('exec', None) or ('excp', traceback)"""
        try:
                exec co in env
        except:
                return ('excp', compact_traceback ())
                
        else:
                return ('exec', None)


def python_prompt (line, env):
        """try eval first, if that fails try exec, return ('eval', result) 
        ('exec', None) or ('excp', traceback)"""
        try:
                
                try:
                        co = compile (line, 'python_line', 'eval')
                except SyntaxError:
                        co = compile (line, 'python_line', 'exec')
                        method, result = python_exec (co, env)
                else:
                        method, result = python_eval (co, env)
        except:
                return ('excp', compact_traceback ())
                
        else:
                return (method, result)
        

# Synopsis
#
# >>> from allegra import prompt
# >>> env = {}
# >>> prompt.python_prompt ('1+1', env)
# ('eval', 2)
# >>> prompt.python_prompt ('a=1+1', env)
# ('exec', None)
# >>> env['a']
# 2
# >>> prompt.python_prompt ('foobar', env)
# ('excp', (
#     'exceptions.NameError', 
#     "name 'foobar' is not defined", 
#     [
#         ('prompt.py', 'python_eval', '53'),
#         ('python_line', '?', '0')
#         ]
#     ))
# >>> try:
# ...     foobar
# ... except:
# ...     prompt.compact_traceback ()
# ...
# (
#     'exceptions.NameError', 
#     "name 'foobar' is not defined", 
#     [('<stdin>', '?', '2')]
#     )



# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/producer/"


import types


class File (object):
        
        "producer wrapper for file[-like] objects"

        def __init__ (self, file, chunk=1<<14): # 16KB buffer
                self.file = file
                self.chunk = chunk

        def more (self):
                return self.file.read (self.chunk)

        def producer_stalled (self):
                return False
        

class Simple (object):
        
        "scanning producer for a large string"

        def __init__ (self, data, chunk=1<<14): # 16KB buffer
                lb = len (data)
                self.content_length = lambda: lb
                self.more = self.produce (data, chunk).next

        def produce (self, data, chunk):
                lb = len (data)
                start = 0
                while start < lb:
                        end = start + chunk
                        yield data[start:end]

                        start = end
                del data, self.content_length, self.more
                yield ''
        
        def producer_stalled (self):
                return False


class Stalled_generator (object):
        
        # the simplest stallable generator, a usefull construct for any
        # generator based producer that is set as a finalization or a 
        # handler of diverse asynchronous or synchronized callback ...
        
        def __call__ (self, *args):
                self.generator = iter ((
                        'Stalled_generator.__call__ not implemented', 
                        ))
        
        generator = None
        
        def more (self):
                try:
                        return self.generator.next ()
                
                except StopIteration:
                        return ''
        
        def producer_stalled (self):
                return self.generator == None
        

class Composite (object):
	
	# This is a more "modern" composite producer than the original
	# one, with support for stalled producers and generators. it is the
	# bread & butter of Allegra's PRESTo! with the Buffer.
	
        def __init__ (self, head, body, glob=1<<14): # 16KB globber
        	assert (
        		type (head) == types.StringType and 
        		type (body) == types.GeneratorType
        		)
        	self.current = head
                self.generator = body
                self.glob = glob
                
        def more (self):
        	if self.current == '':
        		return ''
        		
        	buffer = ''
        	limit = self.glob
        	while True:
	        	if type (self.current) == str:
	        		buffer += self.current
	        		try:
	  				self.current = self.generator.next ()
		  		except StopIteration:
		  			self.current = ''
		  			break
		  		
		  		if len (buffer) > limit:
		  			break
		  			
			elif self.current.producer_stalled ():
				assert buffer != '' # watch this!
				break
				
			else:
				data = self.current.more ()
				if data:
					buffer += data
					if len (buffer) > limit:
						break
					
					else:
						continue
						
	        		try:
	  				self.current = self.generator.next ()
		  		except StopIteration:
		  			self.current = ''
		  			break
		  		
		return buffer
                
	def producer_stalled (self):
		try:
			return self.current.producer_stalled ()
			
		except:
			return False
			
	# Note that this class also makes the original Medusa's lines, buffer 
	# and globbing producer redundant. What this class does it to glob
	# as much strings as possible from a MIME like data structure:
	#
	#	head = 'string'
	#	body = (generator of 'string' or producer ())
	#
        # It's a practical producer for asynchronous REST responses composed
        # of simple strings and maybe-stalling producers. The overhead of
        # another loop buys globbing and helps the peer fill its channel's
        # buffers more efficiently for TCP/IP.


class Tee (object):
        
        def __init__ (self, producer):
                self.index = -1
                self.producer = producer
                try:
                        self.buffers = producer.tee_buffers 
                except AttributeError:
                        self.buffers = producer.tee_buffers = []
                
        def more (self):
                self.index += 1
                try:
                        return self.buffers[self.index]
                
                except IndexError:
                        data = self.producer.more ()
                        self.buffers.append (data)
                        return data
                
        def producer_stalled (self):
                if self.index + 1 < len (self.buffers):
                        return False
                
                return self.producer.producer_stalled ()



# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_server/"

import socket, time

try:
        SOCKET_FAMILIES = (socket.AF_INET, socket.AF_UNIX)
except:
        SOCKET_FAMILIES = (socket.AF_INET, )

from allegra import async_loop, async_core, async_limits


class Listen (async_core.Dispatcher):
        
        server_when = 0.0
        ac_in_meter = ac_out_meter = server_dispatched = 0
        
        def __init__ (
                self, Dispatcher, addr, precision, max, 
                family=socket.AF_INET
                ):
                assert (
                        type (precision) == int and precision > 0 and
                        type (max) == int and max > 0 and
                        family in SOCKET_FAMILIES
                        )
                self.server_dispatchers = []
                self.server_named = {}
                self.Server_dispatcher = Dispatcher
                self.server_precision = precision
                self.create_socket (family, socket.SOCK_STREAM)
                self.set_reuse_addr ()
                self.bind (addr)
                self.listen (max)
                anonymous (self)
                accept_all (self)
                metered (self)             
                self.log ('listen %r' % (addr,), 'info')

        def __repr__ (self):
                return 'async-server id="%x"' % id (self)

        def readable (self):
                return self.accepting

        def writable (self):
                return False

        def handle_accept (self):
                try:
                        conn, addr = self.accept ()
                except socket.error:
                        assert None == self.log (
                                'accept-bogus-socket', 'error'
                                )
                        return
                        #
                        # Medusa original comments from Sam Rushing:
                        #
                        # linux: on rare occasions we get a bogus socket back 
                        # from accept. socketmodule.c:makesockaddr complains 
                        # that the address family is unknown. We don't want 
                        # the whole server to shut down because of this.
                
                except TypeError:
                        assert None == self.log (
                                'accept-would-block', 'error'
                                )
                        return
                        #
                        # Medusa original comments from Sam Rushing:
                        #
                        # unpack non-sequence.  this can happen when a read 
                        # event fires on a listening socket, but when we call 
                        # accept() we get EWOULDBLOCK, so dispatcher.accept() 
                        # returns None. Seen on FreeBSD3.

                name = self.server_resolved (addr)
                if name != None:
                        try:
                                self.server_named[name] += 1
                        except KeyError:
                                self.server_named[name] = 1
                        if self.server_accepted (conn, addr, name):
                                self.server_accept (
                                        conn, addr, name
                                        )
                        else:
                                conn.close ()
                        return
                
                if self.server_resolve == None:
                        self.server_unresolved (conn, addr)
                        conn.close ()
                        return
                        
                def resolve (name):
                        try:
                                self.server_named[name] += 1
                        except KeyError:
                                self.server_named[name] = 1
                        if (
                                name == None and 
                                self.server_unresolved (conn, addr)
                                ):
                                conn.close ()
                        elif self.server_accepted (conn, addr, name):
                                self.server_accept (
                                        conn, addr, name
                                        )
                self.server_resolve (addr, resolve)
                        
        def handle_close (self):
                "close all dispatchers, close the server and finalize it"
                for dispatcher in tuple (self.server_dispatchers):
                        dispatcher.handle_close ()
                self.server_stop (time.time ())
                self.close ()
                self.__dict__ = {}
                #
                # Breaks any circular reference through attributes, by 
                # clearing them all. Note that this prevents finalizations
                # to be used with listeners, but anyway subclassing the
                # stop and shutdown methods provides enough leverage to
                # gracefully come to closure: the listener is most probably 
                # in the highest level of the application's instance tree.
                #
                # ... and sometimes damn hard to finalize ;-)
                
        def server_unresolved (self, conn, addr):
                assert None == self.log ('unresolved %r' % (addr,), 'debug')
                return False # don't care!

        def server_accept (self, conn, addr, name):
                assert None == self.log ('accepted %r' % (addr,), 'debug')
                now = time.time ()
                dispatcher = self.Server_dispatcher ()
                dispatcher.set_connection (conn, addr)
                dispatcher.server_name = name
                dispatcher.server_when = now
                dispatcher.async_server = self
                self.server_decorate (dispatcher, now)
                if self.server_when == 0:
                        self.server_start (now)
                self.server_dispatchers.append (dispatcher)
                return dispatcher
                
        def server_start (self, when):
                "handle the client management startup"
                self.server_when = when
                async_loop.schedule (
                        when + self.server_precision, self.server_manage
                        )
                assert None == self.log ('start', 'debug')
                  
        def server_manage (self, when):
                if not self.server_dispatchers:
                        if self.accepting:
                                self.server_stop (when)
                        else:
                                self.handle_close ()
                        return
                
                if self.server_limit != None:
                        for dispatcher in tuple (self.server_dispatchers):
                                if self.server_limit (dispatcher, when):
                                        self.server_overflow (dispatcher)
                return (when + self.server_precision, self.server_manage)
        
        def server_overflow (self, dispatcher):
                "assert debug log and close an overflowed dispatcher"
                assert None == dispatcher.log ('limit overflow', 'debug')
                dispatcher.handle_close ()
                  
        def server_meter (self, dispatcher):
                "assert debug log and account I/O meters of a dispatcher"
                assert None == dispatcher.log (
                        'in="%d" out="%d"' % (
                                dispatcher.ac_in_meter, 
                                dispatcher.ac_out_meter
                                ),  'debug'
                        )
                self.ac_in_meter += dispatcher.ac_in_meter
                self.ac_out_meter += dispatcher.ac_out_meter
                self.server_dispatched += 1

        def server_close (self, dispatcher):
                "remove the dispatcher from list and meter dispatched"
                name = dispatcher.server_name
                if self.server_named[name] > 1:
                        self.server_named[name] -= 1
                else:
                        del self.server_named[name]
                self.server_dispatchers.remove (dispatcher)
                self.server_meter (dispatcher)
                dispatcher.async_server = None
                
        def server_stop (self, when):
                "handle the server scheduled or inpromptu stop"
                if self.server_when:
                        self.log (
                                'stop dispatched="%d"'
                                ' seconds="%f" in="%d" out="%d"' % (
                                        self.server_dispatched,
                                        (when - self.server_when),
                                        self.ac_in_meter,
                                        self.ac_out_meter
                                        ), 'info')
                self.server_when = 0.0
                self.server_dispatched = \
                        self.ac_in_meter = self.ac_out_meter = 0

        def server_shutdown (self):
                "stop accepting connections, close all current when done"
                self.log ('shutdown', 'info')
                if self.server_when:
                        self.accepting = False
                        for dispatcher in tuple (self.server_dispatchers):
                                dispatcher.close_when_done ()
                else:
                        self.handle_close ()
                return True
                        

def anonymous (listen):
        "allways resolved to the empty string"
        listen.server_resolved = (lambda addr: '')
        listen.server_resolve = None
        return listen

def accept_all (listen):
        listen.server_accepted = (lambda conn, addr, name: True)
        return listen

def accept_named (listen, limit):
        def accepted (conn, addr, name):
                if listen.server_named[name] <= limit:
                        return True
                
                if listen.server_named[name] > 1:
                        listen.server_named[name] -= 1
                else:
                        del listen.server_named[name]
                assert None == listen.log (
                        'accept-limit ip="%s"' % name,
                        'error'
                        )
                return False
        
        listen.server_accepted = accepted 
        return listen


def meter (dispatcher, when):
        "decorate a server dispatcher with stream meters"
        async_limits.meter_recv (dispatcher, when)
        async_limits.meter_send (dispatcher, when)
        def close ():
                del (
                        dispatcher.recv, 
                        dispatcher.send, 
                        dispatcher.close
                        )
                dispatcher.close ()
                dispatcher.async_server.server_close (dispatcher)
                
        dispatcher.close = close

def metered (listen, timeout=1<<32):
        "meter I/O for server streams"
        def decorate (dispatcher, when):
                meter (dispatcher, when)
                dispatcher.limit_inactive = timeout
                
        listen.server_decorate = decorate
        listen.server_inactive = timeout
        listen.server_limit = None
        return listen


def inactive (listen, timeout):
        "meter I/O and limit inactivity for server streams"
        assert type (timeout) == int and timeout > 0
        def decorate (dispatcher, when):
                meter (dispatcher, when)
                dispatcher.limit_inactive = listen.server_inactive
                
        listen.server_decorate = decorate
        listen.server_inactive = timeout
        listen.server_limit = async_limits.inactive
        return listen


def limited (listen, timeout, inBps, outBps):
        "throttle I/O and limit inactivity for managed client streams"
        assert (
                type (timeout) == int and timeout > 0 and
                type (inBps ()) == int and inBps () > 0 and
                type (outBps ()) == int and outBps () > 0
                )
        def throttle (dispatcher, when):
                "decorate a client dispatcher with stream limits"
                async_limits.meter_recv (dispatcher, when)
                async_limits.meter_send (dispatcher, when)
                dispatcher.limit_inactive = timeout
                async_limits.throttle_readable (
                        dispatcher, when, listen.ac_in_throttle_Bps
                        )
                async_limits.throttle_writable (
                        dispatcher, when, listen.ac_out_throttle_Bps
                        )
                def close ():
                        del (
                                dispatcher.recv, 
                                dispatcher.send, 
                                dispatcher.readable,
                                dispatcher.writable,
                                dispatcher.close
                                )
                        dispatcher.close ()
                        dispatcher.async_server.server_close (dispatcher)
                        
                dispatcher.close = close

        listen.server_decorate = throttle
        listen.ac_in_throttle_Bps = inBps
        listen.ac_out_throttle_Bps = outBps
        listen.server_limit = async_limits.limit
        return listen


def rationed (listen, timeout, inBps, outBps):
        "ration I/O and limit inactivity for managed client streams"
        assert (
                type (timeout) == int and timeout > 0 and
                type (inBps) == int and inBps > 0 and
                type (outBps) == int and outBps > 0
                )
        listen.ac_in_ration_Bps = inBps
        listen.ac_out_ration_Bps = outBps
        def throttle_in ():
                return int (listen.ac_in_ration_Bps / max (len (
                        listen.server_dispatchers
                        ), 1))

        def throttle_out ():
                return int (listen.ac_out_ration_Bps / max (len (
                        listen.server_dispatchers
                        ), 1))

        limited (listen, timeout, throttle_in, throttle_out)
        return listen
        


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_net/"

import collections, socket

from allegra import async_core


class NetstringError (Exception): pass

def collect_net (next, buffer, collect, terminate):
        "consume a buffer of netstrings into a stallable collector sink"
        lb = len (buffer)
        if next > 0:
                if next > lb:
                        collect (buffer)
                        return next - lb, '', False # buffer more ...

                if buffer[next] == ',':
                        collect (buffer[:next])
                        if terminate (None):
                                return 0, buffer[next+1:], True # stop now!
                        
                else:
                        raise NetstringError, '3 missing comma'
                
                prev = next + 1
        else:
                prev = 0
        while prev < lb:
                pos = buffer.find (':', prev)
                if pos < 0:
                        if prev > 0:
                                buffer = buffer[prev:]
                        if not buffer.isdigit ():
                                raise NetstringError, '1 not a netstring'
                        
                        return 0, buffer, False # buffer more ...
                        
                try:
                        next = pos + int (buffer[prev:pos]) + 1
                except:
                        raise NetstringError, '2 not a length'
                        
                if next >= lb:
                        collect (buffer[pos+1:])
                        return next - lb, '', False # buffer more
                
                elif buffer[next] == ',':
                        if terminate (buffer[pos+1:next]):
                                return 0, buffer[next+1:], True # stop now!
                        
                else:
                        raise NetstringError, '3 missing comma'
                      
                prev = next + 1 # continue ...
        return 0, '', False # buffer consumed.


class Dispatcher (async_core.Dispatcher):

        ac_in_buffer_size = ac_out_buffer_size = 1 << 14 # sweet 16 kilobytes

        terminator = 0
        collector_stalled = False
                
        def __init__ (self):
                self.ac_in_buffer = ''
                self.ac_out_buffer = ''
                self.output_fifo = collections.deque ()

        def __repr__ (self):
                return 'async-net id="%x"' % id (self)
                
        def readable (self):
                "predicate for inclusion in the poll loop for input"
                return not (
                        self.collector_stalled or
                        len (self.ac_in_buffer) > self.ac_in_buffer_size
                        )

        def writable (self):
                "predicate for inclusion in the poll loop for output"
                return not (
                        (self.ac_out_buffer == '') and
                        not self.output_fifo and self.connected
                        )

        def handle_read (self):
                "try to buffer more input and parse netstrings"
                try:
                        (
                                self.terminator, 
                                self.ac_in_buffer,
                                self.collector_stalled
                                ) = collect_net (
                                        self.terminator, 
                                        self.ac_in_buffer + self.recv (
                                                self.ac_in_buffer_size
                                                ),
                                        self.async_net_collect, 
                                        self.async_net_terminate
                                        )
                except NetstringError, error:
                        self.async_net_error (error)

        def handle_write (self):
                "buffer out a fifo of strings, try to send or close if done"
                obs = self.ac_out_buffer_size
                buffer = self.ac_out_buffer
                fifo = self.output_fifo
                while len (buffer) < obs and fifo:
                        strings = fifo.popleft ()
                        if strings == None:
                                if buffer == '':
                                        self.handle_close ()
                                        return
                                
                                else:
                                        fifo.append (None)
                                        break
        
                        buffer += ''.join ((
                                '%d:%s,' % (len (s), s) for s in strings
                                ))
                if buffer:
                        sent = self.send (buffer[:obs])
                        if sent:
                                self.ac_out_buffer = buffer[sent:]
                        else:
                                self.ac_out_buffer = buffer
                else:
                        self.ac_out_buffer = ''

        # A compatible interface with Async_chat.close_when_done used by
        # Allegra's TCP clients and servers implementation.
                        
        def close_when_done (self):
                """close this channel when previously queued strings have
                been sent, or close now if the queue is empty."""
                if self.output_fifo:
                        self.output_fifo.append (None)
                else:
                        self.handle_close ()

        # The Async_net Interface

        def async_net_out (self, strings):
                "buffer netstrings of an iterable of 8-bit byte strings"
                self.ac_out_buffer += ''.join ((
                        '%d:%s,' % (len (s), s) for s in strings
                        ))

        def async_net_push (self, strings):
                "push an iterable of 8-bit byte strings for output"
                assert hasattr (strings, '__iter__')
                self.output_fifo.append (strings)

        def async_net_pull (self):
                "try to consume the input netstrings buffered"
                if not self.ac_in_buffer:
                        self.collector_stalled = False
                        return
                
                try:
                        (
                                self.terminator, 
                                self.ac_in_buffer,
                                self.collector_stalled
                                ) = collect_net (
                                        self.terminator, 
                                        self.ac_in_buffer,
                                        self.async_net_collect, 
                                        self.async_net_terminate
                                        )
                except NetstringError, error:
                        self.async_net_error (error)

        async_net_in = ''

        def async_net_collect (self, bytes):
                "collect an incomplete netstring chunk into a buffer"
                self.async_net_in += bytes
                
        def async_net_terminate (self, bytes):
                "terminate a collected or buffered netstring and continue"
                if bytes == None:
                        bytes = self.async_net_in
                        self.async_net_in = ''
                return self.async_net_continue (bytes)

        def async_net_continue (self, bytes):
                "assert debug log of collected netstrings"
                assert None == self.log (bytes, 'async-net-continue')
                return False
        
        def async_net_error (self, message):
                "log netstrings error and close the channel"
                self.log (message, 'async-net-error')
                self.handle_close ()
                
                


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

"http://laurentszyster.be/blog/reactor/"


class Buffer (object):

	"a buffer reactor, usefull for asynchronous/synchronous proxy"

        collector_is_simple = True
        buffer_reactor_complete = False

        def __init__ (self):
                self.buffers = []
                self.collect_incoming_data = self.buffers.append
                
	def found_terminator (self):
		self.buffer_reactor_complete = True
                return True

        def more (self):
                try:
                        return self.buffers.pop (0)
                    
                except IndexError:
                        return ''

        def producer_stalled (self):
                return (
			len (self.buffers) == 0 and
			self.buffer_reactor_complete == False
                        )

# The simplest reactor implementation: a pipe


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/synchronized/"

import collections, subprocess

from allegra import loginfo, finalization, thread_loop 


# File producer and collector

def sync_open (self, filename, mode, bufsize):
        try:
                self.sync_file = open (filename, mode, bufsize)
        except:
                self.select_trigger ((self.async_close, ('eo', )))
        else:
                self.select_trigger ((self.async_open, (mode, )))
        
def sync_write (self, data):
        try:
                self.sync_file.write (data)
        except:
                self.select_trigger ((self.async_close, ('ew', )))
        
def sync_read (self):
        try:
                data = self.sync_file.read (self.sync_buffer)
        except:
                self.select_trigger ((self.async_close, ('er', )))
        else:
                if data:
                        self.select_trigger ((self.async_read, (data, )))
                else:
                        sync_close (self, 'r')
        
def sync_close (self, mode):
        try:
                self.sync_file.close ()
        except:
                self.select_trigger ((self.async_close, ('ec', )))
        else:
                self.select_trigger ((self.async_close, (mode, )))
        self.sync_file = None
        

class File_producer (object):

        synchronizer = None
        synchronizer_size = 2
        
        async_buffers = ()
        async_closed = False

        def __init__ (self, filename, mode='rb', bufsize=1<<14):
                assert (
                        type (filename) == str and
                        mode.startswith ('r') and 
                        (0 < len (mode) < 3) and
                        buffer > 0
                        )
                self.sync_buffer = bufsize
                self.async_buffers = collections.deque([])
                thread_loop.synchronize (self)
                self.synchronized ((
                        sync_open, (self, filename, mode, bufsize)
                        ))

        def __repr__ (self):
                return 'synchronized-file-producer id="%x"' % id (self)
                        
        def more (self):
                try:
                        return self.async_buffers.popleft ()
                        
                except:
                        return ''
                        
        def producer_stalled (self):
                return not (
                        self.async_closed or len (self.async_buffers) > 0
                        )
                        
        def async_open (self, mode):
                self.synchronized ((sync_read, (self, )))
        
        def async_read (self, data):
                self.async_buffers.append (data)
                self.synchronized ((sync_read, (self, )))
                
        def async_close (self, mode):
                self.async_closed = True
                thread_loop.desynchronize (self)
                                

class File_collector (object):

        synchronizer = None
        synchronizer_size = 2
        
        collector_is_simple = True
        
        async_closed = False

        def __init__ (self, filename, mode='wb', bufsize=-1):
                assert (
                        type (filename) == str and
                        not mode.startswith ('r') and 
                        (0 < len (mode) < 3)
                        )
                thread_loop.synchronize (self)
                self.synchronized ((
                        sync_open, (self, filename, mode, bufsize)
                        ))

        def __repr__ (self):
                return 'synchronized-file-collector id="%x"' % id (self)
                
        def collect_incoming_data (self, data):
                self.synchronized ((sync_write, (self, data,)))
                
        def found_terminator (self):
                self.synchronized ((sync_close, (self, 'w', )))
                return True

        def async_open (self, mode): pass
        
        def async_close (self, mode):
                self.async_closed = True
                thread_loop.desynchronize (self)


# Subprocess reactor

def sync_popen (self, args):
        try:
                self.subprocess = subprocess.Popen (*args)
        except Exception, error:
                self.select_trigger ((self.async_except, (error, )))
        else:
                self.select_trigger ((self.async_popen, ()))

def sync_stdin (self, data):
        try:
                self.subprocess.stdin.write (data)
        except Exception, error:
                self.select_trigger ((self.async_except, (error, )))
                
def sync_stdout (self):
        exit = self.subprocess.poll ()
        if exit != None:
                sync_wait (self)
                return
        
        try:
                data = self.subprocess.stdout.read (self.sync_buffer)
        except Exception, error:
                self.select_trigger ((self.async_except, (error, )))
        else:
                if data:
                        self.select_trigger ((self.async_stdout, (data, )))
                else:
                        sync_wait (self)
        
def sync_wait (self):
        if self.subprocess == None:
                self.select_trigger ((self.async_return, (None, )))
                return
                
        sub = self.subprocess
        self.subprocess = None
        sub.stdin.close ()
        sub.stdout.close ()
        if sub.stderr:
                sub.stderr.close ()
        self.select_trigger ((self.async_return, (sub.wait (), )))


class Popen_producer (object):
        
        synchronizer = None
        synchronizer_size = 2

        subprocess = async_code = None
        sync_buffer = 1<<16
        
        def __init__ (self):
                self.async_buffers = collections.deque([])
                thread_loop.synchronize (self)
                
        def more (self):
                try:
                        return self.async_buffers.popleft ()
                        
                except:
                        return ''
        
        def producer_stalled (self):
                return (
                        self.async_code == None and
                        len (self.async_buffers) == 0
                        )

        def async_popen (self):
                self.synchronized ((sync_stdout, (self, )))

        def async_stdout (self, data):
                self.async_buffers.append (data)
                self.synchronized ((sync_stdout, (self, )))
                
        def async_stderr (self, data):
                assert None == loginfo.log (
                        'async_error', 'not implemented'
                        )
                
        def async_except (self, error):
                assert None == loginfo.log (str (error), 'debug')
                sync_wait (self)
                
        def async_return (self, code):
                self.async_code = code
                thread_loop.desynchronize (self)
                assert None == loginfo.log ('exit (%r)' % code, 'debug')


def popen_producer (
        args, bufsize=0, executable=None, 
        stdin=subprocess.PIPE, 
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE, 
        preexec_fn=None, close_fds=False, 
        shell=False, cwd=None, env=None, 
        universal_newlines=False, startupinfo=None, 
        creationflags=0
        ):
        assert (
                stdin == subprocess.PIPE and 
                stdout == subprocess.PIPE and
                stderr in (subprocess.PIPE, subprocess.STDOUT)
                )
        sp = Popen_producer ()
        sp.synchronized ((sync_popen, (sp, (
                args, bufsize, executable, stdin, stdout, stderr,
                preexec_fn, close_fds, shell, cwd, env, 
                universal_newlines, startupinfo, creationflags
                ))))
        return sp
        

class Popen_collector (object):
        
        synchronizer = None
        synchronizer_size = 2

        collector_is_simple = True

        subprocess = async_code = None
        
        def __init__ (self):
                thread_loop.synchronize (self)
                
        def collect_incoming_data (self, data):
                self.synchronized ((sync_stdin, (self, data,)))
                
        def found_terminator (self):
                self.synchronized ((sync_wait, (self, )))
                return True
        
        def async_popen (self):
                assert None == loginfo.log ('async_popen', 'debug')

        def async_stderr (self, data):
                assert None == loginfo.log (
                        'async_error', 'not implemented'
                        )
                
        def async_except (self, error):
                assert None == loginfo.log (str (error), 'debug')
                sync_wait (self)
                
        def async_return (self, code):
                self.async_code = code
                thread_loop.desynchronize (self)
                assert None == loginfo.log ('%r' % code, 'debug')


def popen_collector (
        args, bufsize=0, executable=None, 
        stdin=subprocess.PIPE, 
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE, 
        preexec_fn=None, close_fds=False, 
        shell=False, cwd=None, env=None, 
        universal_newlines=False, startupinfo=None, 
        creationflags=0
        ):
        assert (
                stdin == subprocess.PIPE and 
                stdout == subprocess.PIPE and
                stderr in (subprocess.PIPE, subprocess.STDOUT)
                )
        sc = Popen_collector ()
        sc.synchronized ((sync_popen, (sc, (
                args, bufsize, executable, stdin, stdout, stderr,
                preexec_fn, close_fds, shell, cwd, env, 
                universal_newlines, startupinfo, creationflags
                ))))
        return sc


class Popen_reactor (finalization.Finalization):
        
        def __init__ (
                self, args, 
                bufsize=0, executable=None, 
                stdin=subprocess.PIPE, 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE, 
                preexec_fn=None, close_fds=False, 
                shell=False, cwd=None, env=None, 
                universal_newlines=False, startupinfo=None, 
                creationflags=0
                ):
                assert (
                        stdin == subprocess.PIPE and 
                        stdout == subprocess.PIPE and
                        stderr in (subprocess.PIPE, subprocess.STDOUT)
                        )
                self.collector = Popen_collector ()
                self.producer = Popen_producer ()
                self.collector.async_popen = self.async_popen
                self.collector.found_terminator = self.found_terminator
                self.producer.async_return = self.async_return
                self.collector.synchronized ((sync_popen, (self.collector, (
                        args, bufsize, executable, stdin, stdout, stderr,
                        preexec_fn, close_fds, shell, cwd, env, 
                        universal_newlines, startupinfo, creationflags
                        ))))
                
        def async_popen (self):
                self.producer.subprocess = self.collector.subprocess
                self.producer.synchronized ((
                        sync_stdout, (self.producer, )
                        ))
                self.collector.async_popen = None
                
        def found_terminator (self):
                self.collector.found_terminator = None
        
        def async_return (self, code):
                scin = self.collector
                spout = self.producer
                scin.async_code = spout.async_code = code
                self.producer = self.collector = None
                spout.async_return = None
                thread_loop.desynchronize (scin)
                thread_loop.desynchronize (spout)
                assert None == loginfo.log ('%r' % code, 'debug')
        


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

"http://laurentszyster.be/blog/netstring/"


class NetstringsError (Exception): pass


def encode (strings):
        "encode an sequence of 8-bit byte strings as netstrings"
        return ''.join (['%d:%s,' % (len (s), s) for s in strings])

        
def decode (buffer):
	"decode the netstrings found in the buffer, trunk garbage"
	size = len (buffer)
	prev = 0
	while prev < size:
		pos = buffer.find (':', prev)
		if pos < 1:
			break
			
		try:
			next = pos + int (buffer[prev:pos]) + 1
		except:
			break
	
		if next >= size:
			break
		
		if buffer[next] == ',':
			yield buffer[pos+1:next]

		else:
			break
			
		prev = next + 1


def validate (buffer, length):
        "decode the netstrings, but keep garbage and fit to size"
        size = len (buffer)
        prev = 0
        while prev < size and length:
                pos = buffer.find (':', prev)
                if pos < 1:
                        if prev == 0:
                                raise StopIteration # not a netstring!
                        
                        break
                        
                try:
                        next = pos + int (buffer[prev:pos]) + 1
                except:
                        break
        
                if next >= size:
                        break
                
                if buffer[next] == ',':
                        length -= 1
                        yield buffer[pos+1:next]

                else:
                        break
                        
                prev = next + 1

        if length:
                length -= 1
                yield buffer[max (prev, pos+1):]
        
                while length:
                        length -= 1
                        yield ''
                        


def outline (encoded, format, indent):
	"recursively format nested netstrings as a CRLF outline"
	n = tuple (decode (encoded))
	if len (n) > 0:
		return ''.join ((outline (
			e, indent + format, indent
			) for e in n))
			
	return format % encoded


def netstrings (instance):
	"encode a tree of instances as nested netstrings"
	t = type (instance)
	if t == str:
		return instance
		
	if t in (tuple, list, set, frozenset):
		return encode ((netstrings (i) for i in instance))
                
        if t == dict:
                return encode ((netstrings (i) for i in instance.items ()))

	try:
		return '%s' % instance
		
	except:
		return '%r' % instance
		

def netlist (encoded):
	"return a list of strings or [encoded] if no netstrings found"
	return list (decode (encoded)) or [encoded]


def nettree (encoded):
	"decode the nested encoded strings in a tree of lists"
	leaves = [nettree (s) for s in decode (encoded)]
	if len (leaves) > 0:
		return leaves
		
	return encoded


def netlines (encoded, format='%s\n', indent='  '):
	"beautify a netstring as an outline ready to log"
	n = tuple (decode (encoded))
	if len (n) > 0:
		return format % ''.join ((outline (
			e, format, indent
			) for e in n))
			
	return format % encoded


def netoutline (encoded, indent=''):
        "recursively format nested netstrings as an outline with length"
        n = tuple (decode (encoded))
        if len (n) > 0:
                return '%s%d:\n%s%s,\n' % (
                        indent, len (encoded), ''.join ((netoutline (
                                e, indent + '  '
                                ) for e in n)), indent)
        
        return '%s%d:%s,\n' % (indent, len (encoded), encoded)


def netpipe (more, BUFFER_MAX=0):
	"""A practical netstrings pipe generator
	
	Decode the stream of netstrings produced by more (), raise 
	a NetstringsError exception on protocol failure or StopIteration
	when the producer is exhausted.
	
	If specified, the BUFFER_MAX size must be more than twice as
	big as the largest netstring piped through (although netstrings
	strictly smaller than BUFFER_MAX may pass through without raising
	an exception).
	"""
	buffer = more ()
	while buffer:
		pos = buffer.find (':')
		if pos < 0:
			raise NetstringsError, '1 not a netstring' 
		try:
			next = pos + int (buffer[:pos]) + 1
		except:
			raise NetstringsError, '2 not a valid length'
	
		if 0 < BUFFER_MAX < next:
			raise (
				NetstringsError, 
				'4 buffer overflow (%d bytes)' % BUFFER_MAX
				)
			
		while next >= len (buffer):
			data = more ()
			if data:
				buffer += data
			else:
				raise NetstringsError, '5 end of pipe'
		
		if buffer[next] == ',':
			yield buffer[pos+1:next]

		else:
			raise NetstringsError, '3 missing coma'

		buffer = buffer[next+1:]
		if buffer == '' or buffer.isdigit ():
			buffer += more ()

	#
	# Note also that the first call to more must return at least the
	# encoded length of the first netstring, which practically is (or
	# should be) allways the case (for instance, piping in a netstring
	# sequence from a file will be done by blocks of pages, typically
	# between 512 and 4096 bytes, maybe more certainly not less).


if __name__ == '__main__':
        import sys
        assert None == sys.stderr.write (
                'Allegra Netstrings'
                ' - Copyright 2005 Laurent A.V. Szyster'
                ' | Copyleft GPL 2.0\n\n'
                )
        if len (sys.argv) > 1:
        	command = sys.argv[1]
 	else:
		command = 'outline'
 	if command in ('outline', 'decode'):
 		if len (sys.argv) > 2:
	 		if len (sys.argv) > 3:
				try:
					buffer_max = int (sys.argv[3])
				except:
					sys.stderr.write (
						'3 invalid buffer max\n'
						)
					sys.exit (3)
					
			else:
				buffer_max = 0
 			try:
				buffer_more = int (sys.argv[2])
			except:
				sys.stderr.write ('2 invalid buffer size\n')
				sys.exit (2)
				
		else:
			buffer_max = 0
	 		buffer_more = 4096
		def more ():
			return sys.stdin.read (buffer_more)
                count = 0
                try:
        		if command == 'outline':
                                write = sys.stdout.write
        			for n in netpipe (more, buffer_max):
        				write (netlines (n))
                                        count += 1
        		else:
        			for n in netpipe (more, buffer_max):
        				count += 1
                finally:
                        assert None == sys.stderr.write ('%d' % count)
	elif command == 'encode':
                write = sys.stdout.write 
		for line in sys.stdin.xreadlines ():
			write ('%d:%s,' % (len (line)-1, line[:-1]))
	else:
		sys.stderr.write ('1 invalid command\n')
		sys.exit (1)
		
	sys.exit (0)


# Copyright (C) 2005 Laurent A.V. Szyster
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
#    http://www.gnu.org/copyleft/gpl.html
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"http://laurentszyster.be/blog/async_limits/"

import time

from allegra import async_loop


# Metering for stream and datagram sockets

def meter_recv (dispatcher, when):
        "decorate a stream transport with an input meter"
	dispatcher.ac_in_meter = 0
	dispatcher.ac_in_when = when
	metered_recv = dispatcher.recv
	def recv (buffer_size):
		data = metered_recv (buffer_size)
	        dispatcher.ac_in_meter += len (data)
		dispatcher.ac_in_when = time.time ()
	        return data
	        
	dispatcher.recv = recv

def meter_send (dispatcher, when):
        "decorate a stream transport with an output meter"
	dispatcher.ac_out_meter = 0
	dispatcher.ac_out_when = when
	metered_send = dispatcher.send
	def send (data):
		sent = metered_send (data)
	        dispatcher.ac_out_meter += sent
		dispatcher.ac_out_when = time.time ()
	        return sent
	        
	dispatcher.send = send

def meter_recvfrom (dispatcher, when):
        "decorate a datagram transport with an input meter"
        dispatcher.ac_in_meter = 0
        dispatcher.ac_in_when = when
        metered_recvfrom = dispatcher.recvfrom
        def recvfrom (datagram_size):
                data, peer = metered_recvfrom (datagram_size)
                dispatcher.ac_in_meter += len (data)
                dispatcher.ac_in_when = time.time ()
                return data, peer
                
        dispatcher.recvfrom = recvfrom

def meter_sendto (dispatcher, when):
        "decorate a datagram transport with an output meter"
        dispatcher.ac_out_meter = 0
        dispatcher.ac_out_when = when
	metered_sendto = dispatcher.sendto
	def sendto (self):
		sent = metered_sendto (data, peer)
	        dispatcher.ac_out_meter += sent
		dispatcher.ac_out_when = time.time ()
	        return sent

	dispatcher.sendto = sendto


# Inactivity Limits

def inactive_in (dispatcher, when):
        "overflow if connected, not closing and input is inactive"
        return not dispatcher.closing and dispatcher.connected and (
                when - dispatcher.ac_in_when
                ) > dispatcher.limit_inactive
        
def inactive_out (dispatcher, when):
        "overflow if connected, not closing and output is inactive"
        return not dispatcher.closing and dispatcher.connected and (
                when - dispatcher.ac_out_when
                ) > dispatcher.limit_inactive
        
def inactive (dispatcher, when):
        "overflow if connected, not closing and I/O is inactive"
        return not dispatcher.closing and dispatcher.connected and (
                when - max (
                        dispatcher.ac_in_when, dispatcher.ac_out_when
                        )
                ) > dispatcher.limit_inactive
        

# Throttling Decorators

def throttle_readable (dispatcher, when, Bps):
        "decorate a metered dispatcher with an input throttle"
	dispatcher.ac_in_throttle = Bps ()
	dispatcher.ac_in_throttle_when = when
	dispatcher.ac_in_throttle_Bps = Bps
	throttled_readable = dispatcher.readable
	def readable ():
		return (
			dispatcher.ac_in_meter < dispatcher.ac_in_throttle
			and throttled_readable ()
			)
	dispatcher.readable = readable
        
def throttle_writable (dispatcher, when, Bps):
        "decorate a metered dispatcher with an output throttle"
	dispatcher.ac_out_throttle = Bps ()
	dispatcher.ac_out_throttle_when = when
	dispatcher.ac_out_throttle_Bps = Bps
	throttled_writable = dispatcher.writable
	def writable ():
		return (
			dispatcher.ac_out_meter < dispatcher.ac_out_throttle
			and throttled_writable ()
			)
	dispatcher.writable = writable


# Throttling limits

def throttle_in (dispatcher, when):
        "allocate input bandiwth to a throttled dispatcher"
        if dispatcher.ac_in_meter >= dispatcher.ac_in_throttle:
                dispatcher.ac_in_throttle += int ((
                        when - max (
                                dispatcher.ac_in_when,
                                dispatcher.ac_in_throttle_when
                                )
                        ) * dispatcher.ac_in_throttle_Bps ())
        dispatcher.ac_in_throttle_when = when
        return False

        #
        # when the dispatcher exceeded its limit, allocate bandwith at a given
        # rate for the period between "when" - approximatively but steadily
        # "now" - and the last I/O or the last allocation, which ever comes
        # later. in effect it grants the dispatcher the bandwith it is entitled
        # to for the immediate past.
        #
        # the async_throttle_in method is supposed to be called by a
        # periodical defered. for peers with long-lived dispatchers it is
        # faster to periodically allocate bandwith than to do it whenever 
        # we send or receive, or every time we check for readability or 
        # writability.

def throttle_out (dispatcher, when):
        "allocate output bandiwth to a throttled dispatcher"
	if dispatcher.ac_out_meter >= dispatcher.ac_out_throttle:
		dispatcher.ac_out_throttle += int ((
			when - max (
				dispatcher.ac_out_when,
				dispatcher.ac_out_throttle_when
				)
			) * dispatcher.ac_out_throttle_Bps ())
	dispatcher.ac_out_throttle_when = when	
        return False
	

def throttle (dispatcher, when):
        "allocate I/O bandiwth to a throttled dispatcher"
        throttle_in (dispatcher, when)
        throttle_out (dispatcher, when)
        return False


# Limit recurrence factory

def limit_schedule (dispatcher, when, interval, limit, unlimit):
        "instanciate and schedule a limit recurrence"
        # set the limit flag down
        dispatcher.limit_stop = False
        def scheduled (when):
                if dispatcher.closing or dispatcher.limit_stop:
                        # closing or limit flag raised, remove
                        unlimit (dispatcher)
                        return
                
                if limit (dispatcher, when):
                        # limit overflowed, remove and handle close
                        unlimit (dispatcher)
                        dispatcher.handle_close ()
                        return
                
                # recur at interval
                return (when + interval, scheduled)
        
        async_loop.schedule (when + interval, scheduled)

        # I like that one (nested namespaces rule ,-)


# Conveniences: ready-made metering, inactivity check and throttling

def limit_in (dispatcher, when):
        "overflow if input is inactive, throttle it otherwise"
        return (
                inactive_in (dispatcher, when) or 
                throttle_in (dispatcher, when)
                )

def limit_out (dispatcher, when):
        "overflow if output is inactive, throttle it otherwise"
        return (
                inactive_out (dispatcher, when) or 
                throttle_out (dispatcher, when)
                )

def limit (dispatcher, when):
        "overflow if I/O are inactive, throttle them otherwise"
        return (
                inactive (dispatcher, when) or 
                throttle (dispatcher, when)
                )
                
# for stream transport

def limit_recv (dispatcher, interval, timeout, Bps):
        "meter recv and throttle readable, schedule throttling in"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_recv (dispatcher, when)
        throttle_readable (dispatcher, when, Bps)
        limit_schedule (
                dispatcher, when, interval, limit_in, unlimit_recv
                )

def unlimit_recv (dispatcher):
        "unmeter recv and unthrottle readable"
        del (
                dispatcher.recv, 
                dispatcher.readable, 
                dispatcher.ac_in_throttle_Bps 
                )

def limit_send (dispatcher, interval, timeout, Bps):
        "meter send and throttle writable, schedule throttling out"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_send (dispatcher, when)
        throttle_writable (dispatcher, when, Bps)
        limit_schedule (
                dispatcher, when, interval, limit_out, unlimit_send
                )

def unlimit_send (dispatcher):
        "unmeter send and unthrottle writable"
        del (
                dispatcher.send, 
                dispatcher.writable, 
                dispatcher.ac_out_throttle_Bps
                )

def limit_stream (dispatcher, interval, timeout, inBps, outBps):
        "meter and throttle stream I/O, schedule throttling"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_recv (dispatcher, when)
        throttle_readable (dispatcher, when, inBps)
        meter_send (dispatcher, when)
        throttle_writable (dispatcher, when, inBps)
        limit_schedule (
                dispatcher, when, interval, limit, unlimit_stream
                )

def unlimit_stream (dispatcher):
        "unmeter and unthrottle stream I/O"
        del (
                dispatcher.recv, 
                dispatcher.readable,
                dispatcher.ac_in_throttle_Bps,
                dispatcher.send, 
                dispatcher.writable, 
                dispatcher.ac_out_throttle_Bps
                )

# for datagram transport

def limit_recvfrom (dispatcher, interval, timeout, Bps):
        "meter recvfrom and throttle readable, schedule throttling in"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_recvfrom (dispatcher, when)
        throttle_readable (dispatcher, when, Bps)
        limit_schedule (
                dispatcher, when, interval, limit_in, unlimit_recvfrom
                )

def unlimit_recvfrom (dispatcher):
        "unmeter recvfrom and unthrottle readable"
        del (
                dispatcher.recvfrom, 
                dispatcher.readable, 
                dispatcher.ac_in_throttle_Bps
                )

def limit_sendto (dispatcher, interval, timeout, Bps):
        "meter sendto and throttle writable, schedule throttling out"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_sendto (dispatcher, when)
        throttle_writable (dispatcher, when, Bps)
        limit_schedule (
                dispatcher, when, interval, limit_out, unlimit_sendto
                )

def unlimit_sendto (dispatcher):
        "unmeter sendto and unthrottle writable"
        del (
                dispatcher.sendto, 
                dispatcher.writable,
                dispatcher.ac_out_throttle_Bps
                )

def limit_datagram (dispatcher, interval, timeout, inBps, outBps):
        "meter and throttle datagram I/O, schedule throttling"
        when = time.time ()
        dispatcher.limit_inactive = timeout
        meter_recvfrom (dispatcher, when)
        throttle_readable (dispatcher, when, inBps)
        meter_sendto (dispatcher, when)
        throttle_writable (dispatcher, when, inBps)
        limit_schedule (
                dispatcher, when, interval, limit, unlimit_datagram
                )

def unlimit_datagram (dispatcher):
        "unmeter and unthrottle datagram I/O"
        del (
                dispatcher.recvfrom, 
                dispatcher.readable, 
                dispatcher.ac_in_throttle_Bps,
                dispatcher.sendto, 
                dispatcher.writable,
                dispatcher.ac_out_throttle_Bps
                )


# Note about this implementation
#
# other kind of limits - like an absolute limit on the maximum i/o or
# duration per dispatcher - should be implemented in the final class.
#
#
# The Case for Throttling
#
# Asynchat allows to save server's resources by limiting the i/o buffers
# but for a peer on the edges of the network, the bottleneck is bandwith
# not memory. Compare the 16KBps upload limit of a low-end connection with
# the 512MB of RAM available in most PCs those days ... it would take nine
# hours to upload that much data in such small pipe.
#
# It is a basic requirement for a peer to throttle its traffic to a fraction
# of the bandwith generaly available. Because there *are* other applications
# and system functions that need a bit of bandwith, and peer application tend
# to exhaust network resources pretty fast.
#
